// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
  Ejemplo de programa servidor. Atiende peticiones aritmeticas, el protocolo de transporte sera HTTP
  ver http::Transport y el contenido del mensaje sera el resutaldo de un comm::Codec con los
  (x, y, op) -> El resultado sera estos tres componente mas result.

  Para poder probar el sistema de congestion se puede indicar un numero de milisegundos de
  retardo aplicados a cada contestacion.
  
  Los clientes pueden ser: http_kclient.p http_client.p
*/
#include <iostream>

#include <anna/core/core.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>
#include <anna/xml/DocumentMemory.hpp>

#include <anna/comm/comm.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Handler.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/functions.hpp>


#include <anna/test/Communicator.hpp>


using namespace std;

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("http_rserver::MyHandler")
   {
      anna_memset (a_xmlAttributes, 0, sizeof (a_xmlAttributes));      
   }
   ~MyHandler () {
   }

   static const char* className () throw () { return "http_rserver::ReceiverFactory"; }

private:
   struct Attribute { enum _v { ValueOne, ValueTwo, Operator, Result, Time, Max }; };
   
   test::Communicator* a_communicator;
   xml::DocumentMemory a_xmlRequest;
   xml::Node* a_xmlResponse;
   xml::Attribute* a_xmlAttributes [Attribute::Max];
   
   void initialize () throw (RuntimeException);
   void evRequest (ClientSocket&, const http::Request& request) throw (RuntimeException);
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException) {;}
};

class WIMS20ArithmeticServer : public comm::Application {
public:
   WIMS20ArithmeticServer ();
      
private:
   test::Communicator a_communicator;
   ReceiverFactoryImpl <MyHandler> a_receiverFactory;
   comm::ServerSocket* a_serverSocket;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
   xml::Node* asXML (xml::Node* app) const throw ();
};

using namespace std;
using namespace anna::comm;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   WIMS20ArithmeticServer app;

   http::functions::initialize ();

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("http_server", new anna::TraceWriter ("file.trace", 4048000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

WIMS20ArithmeticServer::WIMS20ArithmeticServer () : 
   Application ("http_rserver", "Servidor WIMS20 de operaciones aritmeticas (iRS)", "1.0") 
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que atender peticiones");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccin IP en la que atender");
   commandLine.add ("d", CommandLine::Argument::Mandatory, "Retardo aplicado a la contestacion");
   commandLine.add ("r", CommandLine::Argument::Optional, "Indicador de reuso de direccion", false);
   commandLine.add ("limit", CommandLine::Argument::Mandatory, "% de ocupacion que permitimos");
   commandLine.add ("n", CommandLine::Argument::Optional, "Numero de mensajes a servir", true);
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
   commandLine.add ("timeout", CommandLine::Argument::Optional, "Timeout (ms) del cliente sin enviar peticiones");
   commandLine.add ("quota", CommandLine::Argument::Optional, "Numero de bytes aceptados antes de cerrar el socket");
}

void WIMS20ArithmeticServer::initialize () 
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   int port = cl.getIntegerValue ("p");
   const comm::Device* device = Network::instantiate ().find (Device::asAddress (cl.getValue ("a")));

   comm::TransportFactory* ttff = &http::Transport::getFactory ();

   if (cl.exists ("quota") == true) 
      ttff->setOverQuotaSize (cl.getIntegerValue ("quota"));

   a_serverSocket = new ServerSocket (INetAddress (device, port), cl.exists ("r"), ttff);
   a_serverSocket->setReceiverFactory (a_receiverFactory);
}

void WIMS20ArithmeticServer::run ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   a_communicator.attach (a_serverSocket);
   a_communicator.setDelay ((Millisecond)cl.getIntegerValue ("d"));

   if (cl.exists ("n") == true)
      a_communicator.setMaxMessage (cl.getIntegerValue ("n"));

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   CongestionController::instantiate ().setLimit (cl.getIntegerValue ("limit"));

   if (cl.exists ("timeout") == true)
      a_communicator.setTimeout ((Millisecond)cl.getIntegerValue ("timeout"));

   a_communicator.accept ();
}

xml::Node* WIMS20ArithmeticServer::asXML (xml::Node* app) const 
   throw ()
{
   xml::Node* node = app::Application::asXML (app);
   
   node->createAttribute ("MaxMessage", a_communicator.getMaxMessage ());
   node->createAttribute ("Message", a_communicator.getMessage ());
   
   return node;
}

void MyHandler::initialize ()
   throw (RuntimeException)
{
   a_communicator = app::functions::component <test::Communicator> (ANNA_FILE_LOCATION);
   
   /**
    * Crea el documento XML que usaremos para codificar la respuesta y 
    * pre-localiza los objetos sobre los que tendr que actuar
    */
   a_xmlResponse = new xml::Node ("Response");
   a_xmlAttributes [Attribute::ValueOne] = a_xmlResponse->createAttribute ("ValueOne", 0);
   a_xmlAttributes [Attribute::ValueTwo] = a_xmlResponse->createAttribute ("ValueTwo", 0);
   a_xmlAttributes [Attribute::Operator] = a_xmlResponse->createAttribute ("Operator", 0);
   a_xmlAttributes [Attribute::Result] = a_xmlResponse->createAttribute ("Result", 0);
   a_xmlAttributes [Attribute::Time] = a_xmlResponse->createAttribute ("Time", 0);
}

/*
 * Recibe una peticion
 */
void MyHandler::evRequest (ClientSocket& clientSocket, const http::Request& httpRequest)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyReceiver", "apply", ANNA_FILE_LOCATION));

   if (a_communicator->canContinue (clientSocket) == false)
      return;      
      
   // Extrace el documento XML de la peticin.
   a_xmlRequest.initialize (httpRequest.getBody ());
   
   const xml::Node* request = a_xmlRequest.parse (); 
   
   http::Response* response = allocateResponse ();

   // Antes de contestar espera el tiempo indicado en la configuracin
   a_communicator->delay ();
   
   /**
    * Obtiene los valores de los atributos del documento XML
    */
   const char* op = request->getAttribute ("Operator")->getCStringValue ();
   int v1 = request->getAttribute ("ValueOne")->getIntegerValue ();
   int v2 = request->getAttribute ("ValueTwo")->getIntegerValue ();
      
   /**
    * Establece el valor de los objetos XML pre-localizados que irn en la respuesta
    */
   a_xmlAttributes [Attribute::ValueOne]->setValue (v1);
   a_xmlAttributes [Attribute::ValueTwo]->setValue (v2);
   a_xmlAttributes [Attribute::Operator]->setValue (op);
   
   // Este dato sirve para calcular el tiempo de respuesta del servidor 
   a_xmlAttributes [Attribute::Time]->setValue (request->getAttribute ("Millisecond")->getIntegerValue ());
   
   response->setStatusCode (200);

   switch (*op) {
      case '+': a_xmlAttributes [Attribute::Result]->setValue (v1 + v2); break;
      case '-': a_xmlAttributes [Attribute::Result]->setValue (v1 - v2); break;
      case '*': a_xmlAttributes [Attribute::Result]->setValue (v1 * v2); break;
      case '/':
         if (v2 == 0) {
            response->setStatusCode (400);
            response->setReasonPhrase ("Division por cero");
         }
         else
            a_xmlAttributes [Attribute::Result]->setValue (v1 / v2);
         break;
      default:
         response->setStatusCode (501);
         break;
   }

   /* Observar que establece directamente el documento XML como cuerpo
    * de la respuesta HTTP.
    */    
   if (response->getStatusCode () == 200)
      response->setBody (a_xmlResponse);
   else
      response->clearBody ();      
   
   clientSocket.send (*response);
}
