// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
   Envia una peticion de interseccion sobre HTTP/XML para ihttp_server

   El cliente de esta aplicacion es: ihttp_server.p => Transporte: http::Transport.
*/
#include <iostream>

#include <string.h>

#include <anna/core/core.hpp>
#include <anna/comm/comm.hpp>

#include <anna/app/functions.hpp>

#include <anna/xml/DocumentFile.hpp>
#include <anna/xml/DocumentMemory.hpp>

#include <anna/io/TextReader.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Handler.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/functions.hpp>

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("ihttp_client::MyHandler") {;}

private:
   http::Response a_httpResponse;

   void evRequest (ClientSocket&, const http::Request&) throw (RuntimeException) {;}
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException);
};

class MyCommunicator : public Communicator {
public:
   MyCommunicator () : Communicator (), a_httpRequest ()
   {
      a_httpRequest.setMethod (http::Method::Type::Get);
      a_httpRequest.setURI ("ihttp_client");
   }

private:
   MyHandler a_httpHandler;
   http::Request a_httpRequest;

   void eventReceiveMessage (ClientSocket &, const Message&) throw (RuntimeException);
   void eventBreakConnection (Server* server) throw ();
   void eventStartup () throw (RuntimeException);
};

class IHTTPClient : public anna::comm::Application {
public:
   IHTTPClient ();

   Server* getServer () const throw () { return a_server; }

private:
   MyCommunicator a_communicator;
   Server* a_server;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
};

using namespace std;

bool st_xmlResponse (false);

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   IHTTPClient app;

   http::functions::initialize ();

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug);
      Logger::initialize ("ihttp_client", new TraceWriter ("file.trace", 4096000));

      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }

   return 0;
}

IHTTPClient::IHTTPClient () :
   Application ("ihttp_client", "IHTTPClient", "2.0.0")
{
   CommandLine& commandLine (CommandLine::instantiate ());

   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccion IP Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("xml", CommandLine::Argument::Optional, "Documento XML que contiene la peticion");
   commandLine.add ("nocheck", CommandLine::Argument::Optional, "Indicador para que no comprueba la validaded del HTML", false);
   commandLine.add ("version", CommandLine::Argument::Optional, "Indica la version HTTP usada para enviar");
   commandLine.add ("uri", CommandLine::Argument::Optional, "URI a usar");
   commandLine.add ("m", CommandLine::Argument::Optional, "Metodo HTTP a usar");
   commandLine.add ("xmlresponse", CommandLine::Argument::Optional, "Visualiza la respuesta como un documento XML", false);
   commandLine.add ("overquota", CommandLine::Argument::Optional, "Numero de bytes a recibir sin identificar el protocolo");
   commandLine.add ("c", CommandLine::Argument::Optional, "Valor de la cookie");
}

void IHTTPClient::initialize () 
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   Network& network = Network::instantiate ();

   Host* host = network.find_host ("host000");
   host->assign (network.find (Device::asAddress (cl.getValue ("a"))));

   comm::TransportFactory& ttff = http::Transport::getFactory ();
   if (cl.exists ("overquota") == true) 
      ttff.setOverQuotaSize (cl.getIntegerValue ("overquota"));

   a_server = host->createServer ("http_server", cl.getIntegerValue ("p"), false, &ttff);
}

void IHTTPClient::run ()
   throw (RuntimeException)
{
   a_communicator.accept ();
}

void MyCommunicator::eventStartup ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   const char* filename = (cl.exists ("xml")) ? cl.getValue ("xml"): NULL;

   if (filename == NULL) {
      a_httpRequest.clearBody ();
   }
   else {
      if (cl.exists ("nocheck") == false) {
         xml::DocumentFile xmlDocument;
         xmlDocument.initialize (filename);
         a_httpRequest.setBody (xmlDocument.getContent ());
      }
      else {
         io::TextReader reader;
         reader.open (filename);
         DataBlock body (true);
         const char* line;
   
         while ((line = reader.fetch ()) != NULL)
            body.append (line, anna_strlen (line));
   
         http::Header* header = a_httpRequest.createHeader (http::Header::Type::Connection);
         header->setValue ("   Keep-Alive    ");
   
         a_httpRequest.setBody (body);
      }
   }

   if (cl.exists ("version"))
      a_httpRequest.setVersion (cl.getValue ("version"));

   if (cl.exists ("uri"))
      a_httpRequest.setURI (cl.getValue ("uri"));

   if (cl.exists ("xmlresponse"))
      st_xmlResponse = true;

   if (cl.exists ("m")) {
      http::Method::Type::_v type = http::Method::Type::asEnum (cl.getValue ("m"));

      if (type == http::Method::Type::None) {
         string msg ("Metodo HTTP no reconocido. ");
         msg += http::Method::Type::asList ();
         throw RuntimeException (msg, ANNA_FILE_LOCATION);
      }

      a_httpRequest.setMethod (type);
   }

   if (cl.exists ("c") == true) {
      http::Header* header = a_httpRequest.createHeader ("Cookie");
      header->setValue (cl.getValue ("c"));
      cout << "Con cookie: " << cl.getValue ("c") << " | " << header->asString () << endl << endl;
   }

   // Header requeridos por la plataforma MovilForum para enviar mensajes WapPush
   a_httpRequest.createHeader (http::Header::Type::ContentType)->setValue ("text/xml; charset=\"iso-8859-1\"");
   a_httpRequest.createHeader (http::Header::Type::UserAgent)->setValue ("ANNA.http 1.x");

   string host = anna::functions::asString ("%s:%d", cl.getValue ("a"), cl.getIntegerValue ("p"));
   a_httpRequest.createHeader (http::Header::Type::Host)->setValue (host);
   a_httpRequest.createHeader (http::Header::Type::Authorization)->setValue (" Basic VGlkTGNtOlQxZExjbQ==");
   
   static_cast <IHTTPClient&> (anna::comm::functions::getApp ()).getServer ()->send (a_httpRequest);
}

void MyCommunicator::eventReceiveMessage (ClientSocket& clientSocket, const Message& message)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   if (clientSocket.support (http::Transport::className ()) == false)
      return;

   a_httpHandler.apply (clientSocket, message);
}

void MyCommunicator::eventBreakConnection (Server* server) 
   throw ()
{
   cout << "Perdida conexion con " << server->asString () << endl;
   requestStop ();
}

void MyHandler::evResponse (ClientSocket& clientSocket, const http::Response& response)
   throw (RuntimeException)
{
   const DataBlock& body = response.getBody ();
   
   cout << "HTTP StatusCode: " << response.getStatusCode () << endl;
   cout << "HTTP Text: " << response.getReasonPhrase () << endl;

   if (st_xmlResponse == false)
      cout << "Resultado: \n" << anna::functions::asString (body, 24) << endl << endl;
   else {
      try {
         xml::DocumentMemory xmlDoc;
         xmlDoc.initialize (response.getBody ());
         cout << xmlDoc.getContentAsCString () << endl << endl;
      }
      catch (RuntimeException& ex) {
         ex.trace ();
      }
   }

   app::functions::component <Communicator> (ANNA_FILE_LOCATION)->requestStop ();
}

