// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>

#include <anna/timex/Engine.hpp>

#include <anna/core/util/Millisecond.hpp>

#include "Context.hpp"
#include "Transaction.hpp"

using namespace std;
using namespace anna;
using namespace test;

http4comm::Context::Context (anna::timex::Engine& timeController, const Millisecond& timeout) : 
   anna::timex::Context<S64> ("http4comm", timeController, timeout) 
{;}

void http4comm::Context::create (const S64 id, comm::ClientSocket& clientSocket) 
   throw (RuntimeException)
{   
   http4comm::Transaction* transaction = static_cast <http4comm::Transaction*> (anna::timex::Context <S64>::open (id, 0));

   transaction->setClientSocket (&clientSocket);
   
   LOGINFORMATION (
      string msg ("http4comm::Context::create | Id: ");
      msg += functions::asHexString (id);
      msg += " | ";
      msg += transaction->asString ();
      Logger::information (msg, ANNA_FILE_LOCATION);
   );
}

/**
 * Si el fuera necesaria la instancia de una transacin y no hay ninguna otra marcada como "no-usada" se
 * creara una nueva instancia.
 */
anna::timex::Transaction* http4comm::Context::createTransaction (const int /*classType*/) 
   throw ()
{
   // Si nuestra aplicacin tratara ms de un tipo de transacin tendramos que obtenerla de
   // distintos Recycler, uno para cada tipo de transaccin.
   return a_transactions.create ();
}

/**
 * Cuando la instancia dela transacin no se est usando se marca como no-usada.
 */
void http4comm::Context::releaseTransaction (anna::timex::Transaction* tt) 
   throw ()
{
   // Si nuestra aplicacin tratata ms de un tipo tendramos que obtener el tipo 
   // de la transacin y liberar la transacin en el Recycler adecuado.
   http4comm::Transaction* aux = static_cast <http4comm::Transaction*> (tt);
   a_transactions.release (aux);
}


/**
 * Interpreta el identificador de la transacin como cadena
 */
string http4comm::Context::identifierAsString (const S64& id) const
   throw ()
{
   return functions::asHexString (id);
}

/*
 * Interpreta el contexto de la transacin para obtener su Id de aplicacin.
 */
const S64& http4comm::Context::contextAsIdentifier (const void* cc) const 
   throw ()
{
   return *((S64*) cc);
}
