// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <iostream>

#include <functional>

#include <anna/core/core.hpp>
#include <anna/xml/xml.hpp>

#include <anna/app/Application.hpp>

class Test : public app::Application {
public:
   Test ();
   
private:
   void run () throw (anna::RuntimeException);      
};

using namespace std;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   Test test;

   xml::functions::initialize ();
   
   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug);
      Logger::initialize ("XML", new TraceWriter ("file.trace", 2048000));

      test.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

Test::Test () : 
   app::Application ("testfunctions", "Checking xpath", "1.0") 
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("xml", CommandLine::Argument::Mandatory, "XML document");
   commandLine.add ("xpath", CommandLine::Argument::Mandatory, "XPath condition");
   commandLine.add ("mode", CommandLine::Argument::Optional, "Result type (simple,full,namespace)");
}

void Test::run () 
   throw (RuntimeException)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   xml::DocumentFile xmlDoc;
   xml::Compiler xmlCompiler;
   XPath xpath ("testing");

   xmlDoc.initialize (commandLine.getValue ("xml"));

   Microsecond init, final;

   const char* expression = commandLine.getValue ("xpath");
   const char* smode = NULL;

   int mode = xml::XPath::Mode::Simple;

   if (commandLine.exists ("mode")) {
      smode = commandLine.getValue ("mode");
      if (anna_strstr (smode, "full") != 0)
         mode = xml::XPath::Mode::Full;
      if (anna_strstr (smode, "namespace") != 0)
         mode |= xml::XPath::Mode::Namespace;
   }

   // Several times to check reuse
   for (int i = 0; i < 5; i ++) {
      init = anna::functions::hardwareClock ();
      xpath.apply (xmlDoc, expression, mode);
      final = anna::functions::hardwareClock  ();
      cout << "Time (apply" << i << "): " << final - init << " us" << endl << endl;
      if (xpath.isEmpty () == false) {
         if (i == 0) {
            const xml::Node* node;
            for (Node::const_child_iterator ii = xpath.node_begin (), maxii = xpath.node_end (); ii != maxii; ii ++) 
               cout << xmlCompiler.apply (XPath::node (ii)) << endl;
         }
      }
      else
         cout << "apply: no node matching expression" << endl << endl;
   }

   init = anna::functions::hardwareClock ();
   bool match = xpath.match (xmlDoc, expression, mode);
   final = anna::functions::hardwareClock  ();
   cout << "Time (match): " << final - init << " us" << endl << endl;

   if (match == false)
      cout << "match: no node matching expression" << endl << endl;

   xpath.apply (xmlDoc, expression, mode);
   const xml::Node* node;
   for (Node::const_child_iterator ii = xpath.node_begin (), maxii = xpath.node_end (); ii != maxii; ii ++) 
      cout << xmlCompiler.apply (XPath::node (ii), xml::Compiler::Mode::Compact | xml::Compiler::Mode::NoNamespaces) << endl;
}

