// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_AccessPoint_hpp
#define anna_comm_AccessPoint_hpp

#include <sys/un.h>

#include <anna/core/RuntimeException.hpp>

#include <anna/comm/INetAddress.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Socket;

/**
   Estructura para mantener la informacion de los extremos de un Socket.

   Dependiendo del tipo de Socket puede tener un punto local y/o un punto
   remoto.

   \see Socket
   \warning This should be internally used
*/
class AccessPoint {
public:
  AccessPoint() : a_path(NULL) {;}
  AccessPoint(const std::string& path) : a_path(new std::string(path)) {;}
  AccessPoint(const INetAddress& inetAddress) : a_path(NULL), a_inetAddress(inetAddress) {;}

  ~AccessPoint() { delete a_path;  }

  const INetAddress& getINetAddress() const throw() { return a_inetAddress; }
  const std::string& getPath() const throw() { return *a_path; }

  void clear() throw() { a_inetAddress.clear(); delete a_path; a_path = NULL; }
  bool isNull() const throw() { return a_path == NULL && a_inetAddress.isNull(); }

  bool operator == (const AccessPoint&) const throw();
  AccessPoint& operator = (const INetAddress&) throw();
  AccessPoint& operator = (const std::string& path) throw();
  AccessPoint& operator = (const AccessPoint&) throw();

  void asString(std::string& msg) const throw();

  void asXML(const char* name, xml::Node* parent) const throw(RuntimeException);

  std::string serialize() const throw();

  void translate(const Socket&, sockaddr*&, int& len) throw(RuntimeException);

private:
  const Socket* a_owner;
  INetAddress a_inetAddress;
  std::string* a_path;
  sockaddr_in a_sockaddr_in;
  sockaddr_un a_sockaddr_un;
};

}
}

#endif

