// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_SureTransport_hpp
#define anna_comm_SureTransport_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/DataBlock.hpp>

#include <anna/comm/Transport.hpp>
#include <anna/comm/TransportFactoryImpl.hpp>

namespace anna {

namespace comm {

class Communicator;

/**
   Class to define main transport layer: message analysis received from any source. Protocol
   knows about message semantics when this message is completed: the main problem is find
   out the size.
*/
class SureTransport : public Transport {
public:
  /**
     Destructor.
  */
  ~SureTransport();

  /**
    Class name literal
    \return Class name literal
  */
  static const char* className() throw() { return "anna::comm::SureTransport"; }

  /**
    Returns the transport layer manager associated to this class.
    \return transport layer manager associated to this class.
  */
  static TransportFactory& getFactory() throw() { return st_transportFactory; }

protected:
  static const short int headerSize = (sizeof(short int) + sizeof(int));
  static const short int initTag = 0xaaaa;
  static TransportFactoryImpl <SureTransport> st_transportFactory;

  /**
     Constructor.
  */
  SureTransport();

private:
  DataBlock a_precodec;

  int calculeSize(const DataBlock& dataBlock) throw(RuntimeException);
  const Message* decode(const DataBlock& message) throw(RuntimeException);
  virtual const DataBlock& code(Message&) throw(RuntimeException);

  friend class anna::Allocator <SureTransport>;
};

}
}

#endif
