// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_TransportFactory_hpp
#define anna_comm_TransportFactory_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/NRMutex.hpp>
#include <anna/core/functions.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Transport;
class ClientSocket;

/**
   Interfaz que deben cumplir los gestores de capas de transporte.
*/
class TransportFactory : public NRMutex {
public:
  /**
     Devuelve el nombre logico de esta factoria de capas de transporte.
     \return el nombre logico de esta factoria de capas de transporte.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Devuelve la longitud maxima  que puede contener el buffer intermedio antes de cerrar el socket
     por considerar que no puede sincronizarlo. Un valor 0 indica que no ha sido establecido
     ningun valor maximo, por lo que el valor a usar sera el indicado por
     anna::comm::Transport::DefaultMaxSize.
  */
  int getOverQuotaSize() const throw() { return a_overQuotaSize; }

  /**
     Establece la longitud maxima que puede contener el buffer intermedio antes de cerrar el socket
     por considerar que no puede sincronizarlo.
     \param overQuotaSize longitud maxima que puede contener el buffer intermedio antes de cerrar el socket
     por considerar que no puede sincronizarlo.
  */
  void setOverQuotaSize(const int overQuotaSize) throw() { a_overQuotaSize = overQuotaSize; }

  /**
     Devuelve una cadena con la informacion relevante sobre esta instancia.
     \return una cadena con la informacion relevante sobre esta instancia.
  */
  std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion relevante sobre esta instancia.
     \param parent Nodo XML del que debe depender el documento generado.
     \return un documento XML con la informacion relevante sobre esta instancia.
  */
  xml::Node* asXML(xml::Node* parent) const throw();

protected:
  /**
     Constructor
     \param name Nombre logico de esta factoria de capas de transporte.
  */
  TransportFactory(const char* name) : a_name(name), a_overQuotaSize(0) {;}

  /**
     Crea una instancia del protocolo de transporte asociado a esta factoria.
     \return La instancia de un nuevo protocolo de transporte.
     \warning Cada uno de los transportes obtenidos debera ser liberado invocando a #release.
  */
  virtual Transport* create() throw()  = 0;

  /**
     Libera la instancia del transporte recibida como parametro.
     \param transport Instancia del protocolo de transporte a liberar.
     \warning El transporte recibido como parametro debera haberse obtenido mediante #create.
  */
  virtual void release(Transport* transport) throw()  = 0;

private:
  const std::string a_name;
  int a_overQuotaSize;

  friend class ClientSocket;
};

}
}

#endif
