// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_ThreadData_hpp
#define anna_core_mt_ThreadData_hpp

#ifdef _MT
#include <map>
#include <anna/core/mt/NRMutex.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>
#endif

namespace anna {

/**
   Template que facilita la creacion de datos dependientes del Thread.
   \see anna::Thread

   \param T Clase asociada a este repositorio, debe tener un constructor vacio.
*/
template <typename T> class ThreadData {
public:
  /**
     Contructor.
     \param name Nombre logico de esta template.
  */
  ThreadData(const char* name) : a_name(name) {;}

  /**
     Devuelve los datos asociados al thread desde el que se invoca  a este metodo.
      Si no existen se crean mediante el contructor vacio.
     \return Los datos asociados al thread desde el que se invoca  a este metodo.
  */
  T& get() throw() {
#ifndef _MT
    return a_data;
#else
    Guard guard(a_mutex, "anna::ThreadData");
    const int index = (int) functions::getCurrentThread();
    data_iterator ii = a_datas.find(index);
    T* result = NULL;

    if(ii == a_datas.end()) {
      result = new T;
      a_datas.insert(value_type(index, result));
    } else
      result = ii->second;

    LOGDEBUG(
      std::string msg("anna::ThreadData { Name: ");
      msg += a_name;
      msg += " | ";
      msg += functions::asHexString(anna_ptrnumber_cast(result));
      msg += " }";
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return *result;
#endif
  }

private:
#ifndef _MT
  T a_data;
  const char* a_name;
#else
  typedef typename std::map <int, T*> container;
  typedef typename container::iterator data_iterator;
  typedef typename container::value_type value_type;
  NRMutex a_mutex;
  container a_datas;
  const char* a_name;
#endif
};
}

#endif

