// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_ComponentManager_hpp
#define anna_core_util_ComponentManager_hpp

#include <anna/core/Singleton.hpp>
#include <anna/core/RuntimeException.hpp>

// STL
#include <string>
#include <map>

namespace anna {
namespace xml {
class Node;
}
}

namespace anna {

class Component;

class ComponentManager : public Singleton <ComponentManager> {

public:

  /**
     Returns component instance for the class name provided, <null> if not found.
     \return component instance for the class name provided, <null> if not found.
  */
  Component* find(const char* className) throw();

  /**
     Class XML representation.
     \param parent XML node over which we will put instance information.
     \return XML documentcon with class content.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();


private:

  std::map < std::string /* class name */, Component* > a_components;


  // private constructor
  ComponentManager() {};

  void attach(Component*) throw(RuntimeException);
  void detach(Component*) throw(RuntimeException);


  friend class Singleton <ComponentManager>;
  friend class Component;  // to access attach & detach
};

}

#endif

