// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_codec_basetypes_Time_hpp
#define anna_diameter_codec_basetypes_Time_hpp


// Local
#include <anna/diameter/codec/basetypes/OctetString.hpp>

#include <anna/config/defines.hpp>
//#include <anna.time.Date.h> // TIMESTAMP_OFFSET_NTP1900_OVER_UNIX1970


namespace anna {

namespace diameter {

namespace codec {

class Avp;

namespace basetypes {

/**
* Diameter Time container (NTP timestamp)
*/
class Time : OctetString {

  U32 a_ntpTimestamp; // NTP


  // Only for derived diameter type:
  void updateBasic() throw(anna::RuntimeException);
  void setPrintableString(const char * printableString) throw(anna::RuntimeException);


public:


  // Class-specific ////////////////////////////////////////////////////////////////////////////////////
  //
  /**
  * Type of timestamps (NTP since 0h on 1 January 1900 UTC, UNIX since 0h on 1 January 1970 UTC)
  */
  struct Timestamp {
    enum _v {
      NTP = 0,
      UNIX = 1
    };

    /**
    * Timestamp type description
    * @param v Timestamp type
    * @return Timestamp description
    */
    static const char* asText(const Timestamp::_v v) throw() { // anna_declare_enum is not safe, because labels don't have to match a sequence
      if(v == Timestamp::NTP) return "NTP";
      else if(v == Timestamp::UNIX) return "UNIX";

      return NULL;
    }
  };

  /**
  * Gets the timestamp value
  *
  * @param timestampType Timestamp reference (NTP/Unix)
  *
  * @return timestamp value
  */
  U32 getTimestamp(Timestamp::_v timestampType = Timestamp::NTP) const throw();

  /**
  * Sets the timestamp value
  *
  * @param timestamp timestamp value
  * @param timestampType Timestamp reference (NTP/Unix)
  */
  void setTimestamp(const U32& timestamp, Timestamp::_v timestampType = Timestamp::NTP) throw();
  //
  //////////////////////////////////////////////////////////////////////////////////////////////////////

  // gets

  std::string getFormatName() const throw() { return "Time"; }


  // helpers

  std::string asPrintableString() throw(anna::RuntimeException) {
    return anna::functions::asString(a_ntpTimestamp);
  }

  std::string asString() throw(anna::RuntimeException);


  // sets

  void decode(const char* buffer, const int size) throw(anna::RuntimeException);


  // exports /////////////////////////////
  using AvpData::getSize;
  using AvpData::code;
  //using OctetString::asPrintableString;
  using AvpData::asDataBlockString;
  //using OctetString::asString;
  using AvpData::asHexString;
  //using OctetString::decode;
  using AvpData::fromPrintableString;
  using AvpData::fromHexString;
};

}
}
}
}

#endif

