// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_codec_basetypes_Unsigned64_hpp
#define anna_diameter_codec_basetypes_Unsigned64_hpp


// Local
#include <anna/diameter/codec/basetypes/AvpData.hpp>

#include <anna/core/util/defines.hpp>


namespace anna {

namespace diameter {

namespace codec {

class Avp;

namespace basetypes {

/**
* Diameter Unsigned64 container
*/
class Unsigned64 : public AvpData {

  U64 a_value;


  // Mandatory for basic diameter type:
  void codeBasic(char* buffer, int &size) throw(anna::RuntimeException);
  void setPrintableString(const char * printableString) throw(anna::RuntimeException);

public:


  // Class-specific ////////////////////////////////////////////////////////////////////////////////////
  //
  /**
  * Gets the Unsigned64 value
  *
  * @return Unsigned64 value
  */
  const U64& getValue() const throw() { return a_value; }

  /**
  * Sets the Unsigned64 value
  *
  * @param value Unsigned64 value
  */
  void setValue(const U64& value) throw() { a_value = value; }
  //
  //////////////////////////////////////////////////////////////////////////////////////////////////////


  // gets

  virtual std::string getFormatName() const throw() { return "Unsigned64"; }
  virtual int getSize() const throw() { return 8; }


  // helpers

  std::string asPrintableString() throw(anna::RuntimeException) {
    return anna::functions::asString((anna::U64)a_value);
  }

  std::string asString() throw(anna::RuntimeException) {
    return asPrintableString();
  }


  // sets

  void decode(const char* buffer, const int size) throw(anna::RuntimeException);
};

}
}
}
}

#endif
