// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_timex_Transaction_hpp
#define anna_timex_Transaction_hpp

#include <anna/core/functions.hpp>

#include <anna/timex/TimeEvent.hpp>
#include <anna/timex/Meter.hpp>

namespace anna {

namespace timex {

/**
   Clase base para las transacciones. Se activa una unica vez y cuando transcurren los milisegundos
   establecidos por TimeEvent::getTimeout se invoca al metodo #expire. La implementacion de este metodo
   sera particular para cada uno de las transacciones.

   Incorpora capacidades que facilitan evaluar la duracin media de cada
   transaccin, que son usadas directamente desde anna::timex::Context, aunque podran ser usadas
   por cualquier otro cdigo.

   @see Clock
   @see Timer
   \see Context
*/
class Transaction : public TimeEvent, public Meter {
public:
  /**
     Constructor.
  */
  Transaction() : a_context(NULL) {;}

  /**
     Contructor.
     \param id Identificador de este evento.
     \param timeout La duracion de este evento indicada en milisegundos.
  */
  Transaction(const Id id, const Millisecond & timeout) :
    TimeEvent(id, timeout),
    a_context(NULL)
  {;}

  /**
     Contructor.
     \param id Identificador de este evento.
     \param timeout La duracion de este evento indicada en milisegundos.
     \param context Contexto asociado a esta transaccion.
  */
  Transaction(const Id id, const Millisecond & timeout, void* context) :
    TimeEvent(id, timeout),
    a_context(context)
  {;}

  /**
     Devuelve el contexto asociado a esta transaccion.
     \return El contexto asociado a esta transaccion.
  */
  void* getContext() throw() { return a_context; }

  /**
     Devuelve el contexto asociado a esta transaccion.
     \return El contexto asociado a esta transaccion.
  */
  const void* getContext() const throw() { return a_context; }

  /**
     Establece el contexto asociado a esta transaccion.
     \param context Contexto asociado a esta transaccion.
  */
  void setContext(void* context) throw() { a_context = context; }

  /**
     Devuelve una cadena con la informacion referente a este temporizador.
     \return Una cadena con la informacion referenta a este temporizador.
  */
  virtual std::string asString() const
  throw() {
    std::string msg("Transaction { ");
    msg += TimeEvent::asString();
    msg += " | Contexto: ";
    msg += (a_context == NULL) ? "(null)" : functions::asHexString(anna_ptrnumber_cast(a_context));
    return msg += " }";
  }

private:
  void* a_context;

  Transaction(const Transaction&);
};

}
}

#endif


