// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_xml_Attribute_hpp
#define anna_xml_Attribute_hpp

#include <anna/core/Allocator.hpp>
#include <anna/xml/Data.hpp>

namespace anna {

namespace xml {

class Namespace;

/**
   Atributo de nodo XML.

   Cada nodo XML puede tener una serie indeterminada de atributos. Por ejemplo:

   \code
      <broadcast>
         <INetAddress Address="204.152.65.15" Port="2000"/>
         <INetAddress Address="204.152.65.47" Port="2002"/>
      </broadcast>
   \endcode

   Cada uno de los nodos \em INetAddress tiene dos atributos (\em Address y \em Port).
*/
class Attribute : public Data {
public:
  /**
     Devuelve el nombre de este atributo.
     \return Nombre de este atributo.
  */
  const char* getName() const throw() { return a_name.c_str(); }

  /**
   * Devuelve el namespace asociado a este atributo. Puede ser NULL.
   * \return el namespace asociado a este atributo. Puede ser NULL.
   */
  const Namespace* getNamespace() const throw() { return a_namespace; }

  /**
   * Devuelve el nombre del nodo.
   * \return El nombre del nodo.
   */
  const std::string& getNameAsString() const throw() { return a_name; }

  /**
     Devuelve una cadena con toda la informacion relevante de esta instancia.
     \return Una cadena con toda la informacion relevante de esta instancia.
  */
  std::string asString() const throw();

private:
  std::string a_name;
  const Namespace* a_namespace;

  /* Para evitar que se pueda crear desde el exterior */
  Attribute() : Data(), a_namespace(NULL) {;}

  void setName(const char* name) throw() { a_name = name; }
  void setNamespace(const Namespace* _namespace) throw() { a_namespace = _namespace; }

  friend class Node;
  friend class Allocator<Attribute>;
};

}
}

#endif
