// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <strings.h>
#include <sys/socket.h>

#include <anna/xml/Node.hpp>

#include <anna/comm/AccessPoint.hpp>
#include <anna/comm/Socket.hpp>
#include <anna/comm/Device.hpp>

using namespace std;
using namespace anna;

bool comm::AccessPoint::operator == (const comm::AccessPoint& rhs) const
throw() {
  if(rhs.isNull() && this->isNull())
    return true;

  if(a_path != NULL && rhs.a_path != NULL)
    return *a_path == *rhs.a_path;

  if(a_inetAddress.isNull() == false && rhs.a_inetAddress.isNull() == false)
    return a_inetAddress == rhs.a_inetAddress;

  return false;
}

comm::AccessPoint& comm::AccessPoint::operator = (const INetAddress & inetAddress)
throw() {
  if(a_path != NULL) {
    delete a_path;
    a_path = NULL;
  }

  a_inetAddress = inetAddress;
  return *this;
}

comm::AccessPoint& comm::AccessPoint::operator = (const std::string & path)
throw() {
  if(a_path == &path)
    return *this;

  a_inetAddress.clear();

  if(a_path == NULL)
    a_path = new std::string(path);
  else
    *a_path = path;

  return *this;
}

comm::AccessPoint& comm::AccessPoint::operator = (const AccessPoint & rhs)
throw() {
  if(this == &rhs)
    return *this;

  if(rhs.isNull()) {
    a_inetAddress.clear();

    if(a_path != NULL) {
      delete a_path;
      a_path = NULL;
    }

    return *this;
  }

  return (rhs.a_path != NULL) ? operator= (*rhs.a_path) : operator= (rhs.a_inetAddress);
}

void comm::AccessPoint::translate(const comm::Socket& socket, sockaddr*& s, int& len)
throw(RuntimeException) {
  switch(socket.getDomain()) {
  case Socket::Domain::Inet:

    if(a_inetAddress.isNull()  == true) {
      string msg(socket.asString());
      msg += " | Cannot attach INET socket without address";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    anna_memset(&a_sockaddr_in, 0, sizeof(a_sockaddr_in));
    a_sockaddr_in.sin_family = AF_INET;
    a_sockaddr_in.sin_port = htons(a_inetAddress.getPort());
    a_sockaddr_in.sin_addr.s_addr = (in_addr_t) a_inetAddress.getDevice()->getAddress();
    s = (sockaddr*) & a_sockaddr_in;
    len = sizeof(a_sockaddr_in);
    break;
  case Socket::Domain::Unix:

    if(a_path  == NULL) {
      string msg(socket.asString());
      msg += "Cannot attach UNIX socket without file";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    s = (sockaddr*) & a_sockaddr_un;
    bzero(s, len = sizeof(a_sockaddr_un));
    a_sockaddr_un.sun_family = AF_UNIX;
    anna_strcpy(a_sockaddr_un.sun_path, a_path->c_str());
    break;
  default:
    s = NULL;
    len = 0;

    if(true) {
      string msg(socket.asString());
      msg += "Domain type unsupported at this version";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    break;
  }
}

void comm::AccessPoint::asString(std::string& msg) const
throw() {
  msg += "{ ";

  if(a_inetAddress.isNull() == false)
    msg += a_inetAddress.asString();
  else if(a_path != NULL) {
    msg += "File: ";
    msg += *a_path;
  } else
    msg += "<null>";

  msg += "  }";
}

void comm::AccessPoint::asXML(const char* name, xml::Node* parent) const
throw(RuntimeException) {
  xml::Node* accessPoint  = parent->createChild(name);

  if(a_inetAddress.isNull() == false)
    a_inetAddress.asXML(accessPoint);
  else if(a_path != NULL)
    accessPoint->createAttribute("File", *a_path);
}

std::string comm::AccessPoint::serialize() const
throw() {
  std::string result;

  if(a_inetAddress.isNull() == false)
    result = a_inetAddress.serialize();
  else if(a_path != NULL)
    result = *a_path;
  else
    result = "unknown";

  return result;
}

