// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/TraceMethod.hpp>

#include <anna/xml/Node.hpp>

#include <anna/comm/Host.hpp>
#include <anna/comm/Server.hpp>
#include <anna/comm/Communicator.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/comm/Device.hpp>
#include <anna/comm/ServerAllocator.hpp>

using namespace std;
using namespace anna;

comm::Host::~Host() {
  for(server_iterator ii = server_begin(), maxii = server_end(); ii != maxii; ii ++)
    delete server(ii);

  a_servers.clear();
  a_devices.clear();
}

int comm::Host::SortBy::value(const Server* server)
throw() {
  return server->getRemotePort();
}


const comm::Server* comm::Host::find_server(const int remotePort) const
throw() {
  return a_servers.find(remotePort);
}

comm::Server* comm::Host::find_server(const int remotePort)
throw() {
  return a_servers.find(remotePort);
}


comm::Server* comm::Host::createServer(const string& name, const int remotePort, const bool autoRecovery, comm::TransportFactory* transportFactory, const bool ignoreIncomingMessages, const bool doConnect)
throw(RuntimeException) {
  return add(createServer(ServerAllocator(name, *this, remotePort, autoRecovery, transportFactory, ignoreIncomingMessages)), remotePort, doConnect);
}

//-------------------------------------------------------------------------------
// (1) Acelera las posibilidades principales de busqueda => por nombre y por
// puerto remoto en que esta atendiendo peticiones.
//
// Ojo!! Siempre devuelve un Server aunque posiblemente no este conectado al
// ServerSocket remoto.
//-------------------------------------------------------------------------------
comm::Server* comm::Host::createServer(const ServerAllocator& serverAllocator)
throw(RuntimeException) {
  const int remotePort = serverAllocator.getRemotePort();
  Server* result = serverAllocator.apply();
  result->setIgnoreIncomingMessages(serverAllocator.getIgnoreIncomingMessages());
  return result;
}

void comm::Host::assign(const Device* device)
throw(RuntimeException) {
  if(contains(device) == false)
    a_devices.push_back(device);
}


string comm::Host::asString() const
throw() {
  string result;
  result = "comm::Host { Name: ";
  result += getName();
  result += " | Devices: ";

  if(a_devices.empty() == true)
    result += "(null)";
  else {
    for(const_device_iterator ii = device_begin(), maxii = device_end(); ii != maxii; ii ++) {
      result += device(ii)->asString();
      result += " ";
    }
  }

  return result += " }";
}


xml::Node* comm::Host::asXML(xml::Node* parent) const
throw(RuntimeException) {
  xml::Node* host = parent->createChild("comm.Host");
  xml::Node* node;
  host->createAttribute("Name", getName());
  node = host->createChild("Devices");

  for(const_device_iterator ii = device_begin(), maxii = device_end(); ii != maxii; ii ++)
    device(ii)->asXML(node);

  node = host->createChild("Servers");

  for(const_server_iterator ss = server_begin(), maxss = server_end(); ss != maxss; ss ++)
    server(ss)->asXML(node);

  return host;
}


comm::Server* comm::Host::add(comm::Server* result, const int remotePort, const bool doConnect)
throw() {
  result->a_sequence = a_servers.size();
  a_servers.add(result);

  try {
    LOGINFORMATION(
      string msg("comm::Host::add");
      msg += (doConnect ? "(server creation and connection) | " : "(server creation, but connection not performed now) | ");
      msg += result->asString();
      Logger::information(msg,  ANNA_FILE_LOCATION)
    );

    if(doConnect) result->connect();
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  return result;
}

