// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <stdarg.h>
#include <stdio.h>
#include <iostream>

#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <anna/core/tracing/TraceWriter.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/util/Configuration.hpp>
#include <anna/core/tracing/TraceWriterObserver.hpp>

using namespace std;
using namespace anna;

TraceWriter::TraceWriter() :
  a_maxSize(DefaultMaxKBSize),
  a_stream(-1),
  a_lastTime(0),
  a_observed(false) {
  string file;
  file = "/var/tmp/anna.process.";
  file += functions::asString((int) getpid());
  file += ".log";
  setup(file.c_str(), DefaultMaxKBSize, false);
}

TraceWriter::TraceWriter(const char* fileName, const int maxSize) :
  a_maxSize(DefaultMaxKBSize),
  a_stream(-1),
  a_lastTime(0),
  a_observed(false) {
  setup(fileName, maxSize, false);
}

void TraceWriter::setup(const char* fileName, const Configuration& configuration)
throw() {
  int maxSize = a_maxSize;
  bool clean = true;

  try {
    try {
      maxSize = (configuration.getIntegerValue("Trace", "MaxFileSize") << 10);
    } catch(...) {
    }

    try {
      Logger::setLevel(Logger::asLevel(configuration.getValue("Trace", "Level")));
    } catch(Exception& ex) {
      ex.trace();
    }

    try {
      clean = (configuration.getIntegerValue("Trace", "Clean") == 1);
    } catch(...) {
    }
  } catch(RuntimeException&) {
  }

  setup(fileName, maxSize, clean);
}

void TraceWriter::setup(const char* fileName, const int maxSize, const bool clean)
throw() {
  if(a_stream != -1) {
    if(clean == false) {
      string msg("\nTraces go on at: ");
      msg += fileName;
      msg += '\n';
      write(a_stream, msg.c_str(), msg.length());
    }

    if(a_observed == true) {
      for(observer_iterator ii = observer_begin(), maxii = observer_end(); ii != maxii; ii ++)
        observer(ii)->handleClose(a_stream);
    }

    close(a_stream);
    a_stream = -1;

    if(clean == true) {
      unlink(a_outputFile.c_str());
      unlink(a_outputOldFile.c_str());
    }
  }

  a_outputFile = fileName;
  a_outputOldFile = a_outputFile;
  a_outputOldFile += ".old";

  if(maxSize >= (256 * 1024))
    a_maxSize = maxSize;

  const char* aux = Logger::asString(Logger::getLevel());
  char date [anna::functions::DateTimeSizeString + 7];
  anna_strcpy(date, "- [");
  anna_strcat(anna_strcat(date, getDate()), "]\n");
  int stream = prepareOutput(date);

  if(stream != ErrorStream) {
    const char* aux = "\n--------------------------------------------------------------\n";
    const char* aux2 = "- Current trace level: ";
    write(stream, aux, anna_strlen(aux));
    write(stream, date, anna_strlen(date));
    write(stream, aux2, anna_strlen(aux2));
    aux2 = Logger::asString(Logger::getLevel());
    write(stream, aux2, anna_strlen(aux2));
    string filesize = functions::asString("\n- Max file size : %d Kb", a_maxSize >> 10);
    aux2 = filesize.c_str();
    write(stream, aux2, anna_strlen(aux2));
    write(stream, aux, anna_strlen(aux));
  }
}

void TraceWriter::attach(TraceWriterObserver* observer)
throw() {
  if(observer != NULL) {
    a_observed = true;
    a_observers.push_back(observer);

    if(a_stream != -1)
      observer->handleOpen(a_stream);
  }
}

void TraceWriter::printResume()
throw() {
  cout << "Traces file ..................................: " << a_outputFile << endl;
  cout << "Backup copy  .................................: " << a_outputOldFile << endl;
  cout << "Files size (Kbytes) ..........................: " << (a_maxSize >> 10) << endl;
  cout << "Traces level .................................: " << Logger::asString(Logger::getLevel()) << endl << endl;
}

void TraceWriter::do_write(int level, const char* text, ...)
throw() {
  va_list ap;
  const char* data;
  int size;
  int nbytes;
  GuardNoLog guard(a_mutex);
  DataBlock& dataBlock(Logger::Writer::getDataBlock());
  data = dataBlock.getData();
  size = dataBlock.getMaxSize();

  if(size == 0) {
    dataBlock.allocate(size = 1024);
    data = dataBlock.getData();
  }

  while(true) {
    va_start(ap, text);
    nbytes = vsnprintf(const_cast <char*>(data), size, text, ap);

    if(nbytes >= size) {
      dataBlock.allocate(nbytes + 1);
      data = dataBlock.getData();
      size = dataBlock.getMaxSize();
      continue;
    }

    break;
  }

  va_end(ap);
  const char* aux = Logger::asString(Logger::Level(level));
  char date [anna::functions::DateTimeSizeString + 5];
  date [0] = '[';
  date [1] = 0;
  anna_strcat(anna_strcat(date, getDate()), "] ");
  a_stream = prepareOutput(date);

  if(a_stream == ErrorStream) {
    cerr << Logger::asString(Logger::Level(level)) << " | " << date << " | " << data << endl << endl;
    return;
  }

  bool ok(false);

  if(write(a_stream, date, anna_strlen(date)) > 0)
    if(write(a_stream, aux, anna_strlen(aux)) > 0)
      if(write(a_stream, " | ", 3) > 0)
        if(write(a_stream, data, nbytes) > 0)
          ok = (write(a_stream, "\n\n", 2) > 0);

  if(ok == false && errno != EINTR)
    perror("Cannot save traces");
}

const char* TraceWriter::getDate()
throw() {
  const Microsecond msnow = functions::hardwareClock();

  if(a_lastTime == 0 || (msnow - a_lastTime) > 1000000L) {
    a_lastTime = msnow;
    char date [functions::DateTimeSizeString];
    a_date = functions::asDateTime(functions::second(), date);
  }

  return a_date.c_str();
}

int TraceWriter::prepareOutput(const char* date)
throw() {
  int result = a_stream;

  if(result == ErrorStream)
    return result;

  if(result != -1) {
    struct stat data;
    int r;
    anna_signal_shield(r, stat(a_outputFile.c_str(), &data));

    if(r == -1 || data.st_size > a_maxSize) {
      if(a_observed == true) {
        for(observer_iterator ii = observer_begin(), maxii = observer_end(); ii != maxii; ii ++)
          observer(ii)->handleClose(a_stream);
      }

      anna_signal_shield(r, close(result));
      rename(a_outputFile.c_str(), a_outputOldFile.c_str());
      result = -1;
    }
  }

  if(result == -1) {
    anna_signal_shield(
      result, open(a_outputFile.c_str(), O_RDWR | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)
    );

    if(result == -1) {
      cerr << "Cannot open the file '" << a_outputFile << "': " << strerror(errno) << endl;
      Logger::setLevel(Logger::Error);
      result = ErrorStream;
    } else {
      int r;
      anna_signal_shield(r, fcntl(result, F_GETFL));

      if(r != -1) {
        const int opts = r | O_NONBLOCK;
        anna_signal_shield(r, fcntl(result, F_SETFL, opts));
      }
    }

    if(a_observed == true) {
      for(observer_iterator ii = observer_begin(), maxii = observer_end(); ii != maxii; ii ++)
        observer(ii)->handleOpen(result);
    }
  }

  return result;
}

