// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter/functions.hpp>
#include <anna/diameter/helpers/helpers.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/time/functions.hpp>

// Local
#include <anna/diameter.comm/Session.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/Transport.hpp>
#include <anna/diameter.comm/Response.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/TimerManager.hpp>
#include <anna/diameter.comm/Timer.hpp>

#include <anna/comm/Network.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/Node.hpp>
#include <anna/timex/Engine.hpp>
#include <anna/app/functions.hpp>

// Standard
#include <stdlib.h> // rand()
#include <time.h>



using namespace std;
using namespace anna::diameter;
using namespace anna::diameter::comm;

//static
const anna::Millisecond Session::DefaultTimeout(10000); // Application messages timeout
const int Session::DefaultPort(3868);


Session::Session(const char *className, const char *timerName) : anna::timex::Timer(timerName, (anna::Millisecond)0) /* not assigned */,
  a_className(className),
  a_timeController(NULL),
  a_engine(NULL),
  a_notifyOrphansOnExpiration(true),
  a_actionTimer(NULL),
  a_dpr(ClassCode::ApplicationMessage) { // realmente no es necesario, los Message son por defecto de aplicacion
  initialize();
}

void Session::initialize() throw() {
  a_state = State::Closed;
  a_socketId = 0;
  a_lastIncomingActivityTime = (anna::Millisecond)0;
  a_lastOutgoingActivityTime = (anna::Millisecond)0;
  a_onDisconnect = OnDisconnect::WaitPendings;

  for(int i = ClassCode::Min; i < ClassCode::Max; i ++)
    a_timeouts [i] = DefaultTimeout;
}

//Session::~Session() {;}


void Session::initializeSequences() throw() {
  // Sequences
  //
  //   Hop-by-Hop Identifier
  //      The Hop-by-Hop Identifier is an unsigned 32-bit integer field (in
  //      network byte order) and aids in matching requests and replies.
  //      The sender MUST ensure that the Hop-by-Hop identifier in a request
  //      is unique on a given connection at any given time, and MAY attempt
  //      to ensure that the number is unique across reboots.  The sender of
  //      an Answer message MUST ensure that the Hop-by-Hop Identifier field
  //      contains the same value that was found in the corresponding
  //      request.  The Hop-by-Hop identifier is normally a monotonically
  //      increasing number, whose start value was randomly generated.  An
  //      answer message that is received with an unknown Hop-by-Hop
  //      Identifier MUST be discarded.
  //
  //   End-to-End Identifier
  //      The End-to-End Identifier is an unsigned 32-bit integer field (in
  //      network byte order) and is used to detect duplicate messages.
  //      Upon reboot implementations MAY set the high order 12 bits to
  //      contain the low order 12 bits of current time, and the low order
  //      20 bits to a random value.  Senders of request messages MUST
  //      insert a unique identifier on each message.  The identifier MUST
  //      remain locally unique for a period of at least 4 minutes, even
  //      across reboots.  The originator of an Answer message MUST ensure
  //      that the End-to-End Identifier field contains the same value that
  //      was found in the corresponding request.  The End-to-End Identifier
  //      MUST NOT be modified by Diameter agents of any kind.  The
  //      combination of the Origin-Host (see Section 6.3) and this field is
  //      used to detect duplicates.  Duplicate requests SHOULD cause the
  //      same answer to be transmitted (modulo the hop-by-hop Identifier
  //      field and any routing AVPs that may be present), and MUST NOT
  //      affect any state that was set when the original request was
  //      processed.  Duplicate answer messages that are to be locally
  //      consumed (see Section 6.2) SHOULD be silently discarded.
  srand(::time(NULL) + anna::functions::hash(anna::functions::asString("%s:%d|%d", getAddress().c_str(), getPort(), a_socketId).c_str()));
  a_nextHopByHop = rand();
  a_nextEndToEnd = ((::time(NULL) & 0xFFF) << 20) + (rand() & 0xFFFFF);
}

void Session::sendDPA()
throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "sendDPA", ANNA_FILE_LOCATION));
  anna::DataBlock dpa(true);
  a_engine->readDPA(dpa, a_dpr.getBody()); // Asume that DPA is valid ...

  if(dpa.isEmpty()) {
    LOGWARNING(anna::Logger::warning("This diameter agent defines an empty DPA message. Remote disconnection DPR will be ignored going to the Bound state", ANNA_FILE_LOCATION));
    setState(State::Bound);
    return;
  }

  Message msgDpa;
  msgDpa.setBody(dpa);
  send(&msgDpa);
  LOGWARNING(anna::Logger::warning("DPA has been sent to the peer", ANNA_FILE_LOCATION));
  // Temporizador de proteccion por si el servidor no cierra:
  //    state            event              action           next state
  //    ---------------------------------------------------------------
  //    R-Open           R-Rcv-DPR          R-Snd-DPA        Closing
  //    Closing          Timeout            Error            Closed
  activateActionTimer(anna::diameter::comm::Timer::Type::SessionUnbind);
}

void Session::setState(State::_v state) throw() {
  LOGDEBUG(

  if(state != a_state) {
  std::string msg("Session state change: ");
    msg += asText(a_state);
    msg += " -> ";
    msg += asText(state);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  }
  );
  a_state = state;
}


void Session::activateActionTimer(const anna::diameter::comm::Timer::Type::_v type) throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "activateActionTimer", ANNA_FILE_LOCATION));
  cancelTimer(); // Session timer

  try {
    if(a_actionTimer && a_actionTimer->isActive()) cancelActionTimer();  // no ocurrira

    a_actionTimer = TimerManager::instantiate().createTimer(this, type);
  } catch(anna::RuntimeException& ex) {
    std::string msg = "CAPTURED EXCEPTION during action timer activation (activateActionTimer): ";
    msg += ex.getText();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
  }
}


void Session::cancelActionTimer() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "cancelActionTimer", ANNA_FILE_LOCATION));

  if(a_actionTimer) {
    if(a_actionTimer->isActive()) {
      try {
        TimerManager::instantiate().cancelTimer(a_actionTimer);
      } catch(anna::RuntimeException& ex) {
        ex.trace();
      }
    } else { // por aqui no deberia pasar ...
      LOGDEBUG(anna::Logger::debug("Timer not activated!", ANNA_FILE_LOCATION));
    }

    a_actionTimer = NULL;
  }
}


void Session::activateTimer() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "activateTimer", ANNA_FILE_LOCATION));
  cancelActionTimer();

  try {
    if(a_timeController == NULL)  // Application must created a timex engine
      a_timeController = anna::app::functions::component <anna::timex::Engine> (ANNA_FILE_LOCATION);

    if(isActive()) cancelTimer();

    a_timeController->activate(this);
  } catch(anna::RuntimeException& ex) {
    std::string msg = "CAPTURED EXCEPTION during session timer activation (activateTimer): ";
    msg += ex.getText();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
  }

  timerStarted();
}


void Session::cancelTimer() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "cancelTimer", ANNA_FILE_LOCATION));

  if(isActive()) {
    try {
      if(a_timeController == NULL)  // Application must created a timex engine
        a_timeController = anna::app::functions::component <anna::timex::Engine> (ANNA_FILE_LOCATION);

      a_timeController->cancel(this);
    } catch(anna::RuntimeException& ex) {
      std::string msg = "CAPTURED EXCEPTION during session timer cancellation (cancelTimer): ";
      msg += ex.getText();
      anna::Logger::error(msg, ANNA_FILE_LOCATION);
    }

    timerStopped();
  } else {
    LOGDEBUG(anna::Logger::debug("Timer not activated!", ANNA_FILE_LOCATION));
  }
}


//-------------------------------------------------------------------------
// Se invoca desde diameter::comm::Timer
//-------------------------------------------------------------------------
void Session::expireResponse(diameter::comm::Response* response)
throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "expireResponse", ANNA_FILE_LOCATION));
  bool doUnbind = false;
  bool doRetransmission = false;

  if(response->getClassCode() != ClassCode::Bind) {
    if(response->getRequest()->getOnExpiry() == Message::OnExpiry::Abandon) {
      a_onDisconnect = OnDisconnect::IgnorePendings; // Abandon is not graceful
      doUnbind = true;
    }
    else if(response->getRequest()->getOnExpiry() == Message::OnExpiry::Retransmit) {
      doRetransmission = true;
    }
  } else
    doUnbind = true; // (*)


  try {
    response->setMessage(NULL);
    eventResponse(*response);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  // DPR special case:
  diameter::CommandId cid = response->getRequest()->getCommandId();

  if((cid == helpers::base::COMMANDID__Disconnect_Peer_Request) && (a_state == State::WaitingDPA)) {
    LOGDEBUG(anna::Logger::debug("Expired DPR sent to remote diameter point: local DPR procedure will be ignored going to the Bound state", ANNA_FILE_LOCATION));
    setState(State::Bound);
  }

  if(doRetransmission) {
    diameter::comm::Message *request = const_cast<Message*>(response->getRequest());
    int retries = request->getRetries();
    if (retries > 0) {
      retries--;
      request->setRetries(retries);
      LOGDEBUG
      (
        std::string msg(asString());
        msg += anna::functions::asString(" | Retransmission initiated for request with HopByHop: %u; remaining %d retries", response->getHopByHop(), retries);
        anna::Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      diameter::codec::functions::setPotentiallyReTransmittedMessageBit(*request);
      eventRequestRetransmission(request);
      send(request);
    }
  }

  response_erase(response);

  if(doUnbind) unbind();
}

void Session::finalize() throw() {
  LOGMETHOD(anna::TraceMethod traceMethod("anna::diameter::comm::Session", "finalize", ANNA_FILE_LOCATION));
  setState(State::Closed);
  cancelTimer(); // Session timer
  cancelActionTimer(); // Action timer
  eventPeerShutdown();
///////////////////////////////////////////////////////////////////////
// Notificar la finalizaci�n de las respuestas pendientes de recibir //
///////////////////////////////////////////////////////////////////////
// RFC 3588 - 5.5.4.  Failover and Failback Procedures
//
//      In the event that a transport failure is detected with a peer, it is
//      necessary for all pending request messages to be forwarded to an
//      alternate agent, if possible.  This is commonly referred to as
//      failover.
//
//      In order for a Diameter node to perform failover procedures, it is
//      necessary for the node to maintain a pending message queue for a
//      given peer.  When an answer message is received, the corresponding
//      request is removed from the queue.  The Hop-by-Hop Identifier field
//      is used to match the answer with the queued request.
//
//      When a transport failure is detected, if possible all messages in the
//      queue are sent to an alternate agent with the T flag set.  On booting
//      a Diameter client or agent, the T flag is also set on any records
//      still remaining to be transmitted in non-volatile storage.  An
//      example of a case where it is not possible to forward the message to
//      an alternate server is when the message has a fixed destination, and
//      the unavailable peer is the message's final destination (see
//      Destination-Host AVP).  Such an error requires that the agent return
//      an answer message with the 'E' bit set and the Result-Code AVP set to
//      DIAMETER_UNABLE_TO_DELIVER.
//
//      It is important to note that multiple identical requests or answers
//      MAY be received as a result of a failover.  The End-to-End Identifier
//      field in the Diameter header along with the Origin-Host AVP MUST be
//      used to identify duplicate messages.
  Response* response;

  for(response_iterator ii = response_begin(), maxii = response_end(); ii != maxii; ii ++) {
    response = Session::response(ii);
    response->setResultCode(Response::ResultCode::DiameterUnavailable);

    if(!a_notifyOrphansOnExpiration) {  // to avoid message bursts (to alternate servers for client-session context), we will manage at expireResponse
      response->cancelTimer();

      try {
        response->setMessage(NULL);
        eventResponse(*response);
      } catch(anna::RuntimeException& ex) {
        ex.trace();
      }

      Response::release(response);
    }
  }

  if(a_notifyOrphansOnExpiration) return;

  a_responses.clear();
}

void Session::response_add(Response* response)
throw() {
  a_responses.add(response);
  response->setSession(this);

  try {
    response->activateTimer();
  } catch(anna::Exception& ex) {
    ex.trace();
  }
}

void Session::response_erase(Response* response)
throw() {
  a_responses.erase(response);
  Response::release(response);

  if(a_state == State::Disconnecting)  // only OnDisconnect::WaitPendings arrives here (the other disconnect suddently)
    if(getOTARequests() == 0) sendDPA();

  if(a_state == State::Closing)  // only OnDisconnect::WaitPendings arrives here (the other disconnect suddently)
    if(getOTARequests() == 0) unbind();
}

Response* Session::response_find(const HopByHop hopByHop)
throw(anna::RuntimeException) {
  diameter::comm::Response* result = a_responses.find(hopByHop);
//   if (result == NULL) {
//      string msg(asString());
//      msg += anna::functions::asString(" | Response received for non registered context (HopByHop: %u)", hopByHop);
//      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
//   }
  //Message * message = const_cast<Message*>(result->getRequest());
  return result;
}

std::string Session::asString() const
throw() {
  string result(a_className);
  result += " { ";
  result += anna::timex::Timer::asString();
  result += " | Socket Id: ";
  result += anna::functions::asString(a_socketId);
  result += " | State: ";
  result += asText(a_state);
  result += " | OnDisconnect: ";
  result += asText(a_onDisconnect);
  result += " | Next Hop by hop: ";
  result += anna::functions::asString(a_nextHopByHop);
  result += " | Next End to end: ";
  result += anna::functions::asString(a_nextEndToEnd);
  result += anna::functions::asString(" | OTA requests: %d%s", getOTARequests(), idle() ? " (idle)" : "");
  result += " | Last Incoming Activity Time: ";
  result += a_lastIncomingActivityTime.asString();
  result += " | Last Outgoing Activity Time: ";
  result += a_lastOutgoingActivityTime.asString();

  for(int i = ClassCode::Bind; i < ClassCode::Max; i ++) {
    result += " | Timeout for ClassCode '";
    result += ClassCode::asText((ClassCode::_v)i);
    result += "': ";
    result += a_timeouts[i].asString();
  }

  return result;
}

anna::xml::Node* Session::asXML(anna::xml::Node* parent) const
throw() {
  //parent = anna::timex::Timer::asXML(parent);
  anna::xml::Node* result = parent->createChild("diameter.comm.Session");
  result->createAttribute("SocketId", anna::functions::asString(a_socketId));
  result->createAttribute("State", asText(a_state));
  result->createAttribute("OnDisconnect", asText(a_onDisconnect));
  result->createAttribute("NextHopByHop", anna::functions::asString(a_nextHopByHop));
  result->createAttribute("NextEndToEnd", anna::functions::asString(a_nextEndToEnd));
  result->createAttribute("OTArequests", anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : ""));
  result->createAttribute("LastIncomingActivityTime", a_lastIncomingActivityTime.asString());
  result->createAttribute("LastOutgoingActivityTime", a_lastOutgoingActivityTime.asString());

  for(int i = ClassCode::Bind; i < ClassCode::Max; i ++) {
    std::string name = "TimeoutFor"; name += ClassCode::asText((ClassCode::_v)i);
    result->createAttribute(name.c_str(), a_timeouts[i].asString());
  }

  // Messages
  anna::xml::Node* messages = result->createChild("diameter.comm.Messages");
  const Response* response;
  const Message* message;

  for(const_response_iterator ii = response_begin(), maxii = response_end(); ii != maxii; ii ++) {
    if((message = Session::response(ii)->getRequest()) != NULL)
      message->asXML(messages);
  }

  return result;
}

const char* Session::asText(const State::_v state)
throw() {
  static const char* text [] = { "Closed", "WaitingBind", "Bound", "Failover", "Suspect", "WaitingDPA", "Disconnecting", "Closing" };
  return text [state];
}

const char* Session::asText(const OnDisconnect::_v onDisconnect)
throw() {
  static const char* text [] = { "IgnorePendings", "WaitPendings" };
  return text [onDisconnect];
}



HopByHop Session::SortById::value(const Response* response)
throw() {
  return response->getHopByHop();
}


//------------------------------------------------------------------------------
//---------------------------------------- Session::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void Session::updateIncomingActivityTime() throw() {
  a_lastIncomingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated INCOMING activity on session (milliseconds unix): ";
    msg += anna::functions::asString(a_lastIncomingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}


//------------------------------------------------------------------------------
//---------------------------------------- Session::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void Session::updateOutgoingActivityTime(void) throw() {
  a_lastOutgoingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated OUTGOING activity on session (milliseconds unix): ";
    msg += anna::functions::asString(a_lastOutgoingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}
