// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/app/dcca/Message.hpp>


using namespace anna::diameter::app::dcca;


const qosProfile_t & Message::decode3GPPGPRSNegQoSProfile() throw(anna::RuntimeException) {
  const std::string & utf8 = get3GPPGPRSNegQoSProfile()->getValue();
  a_qosProfile.reset();
  return a_qosProfile;
}


//
//
//      5 - 3GPP-GPRS Negotiated QoS profile
//
//          Bits
//      Octets    8  7  6  5  4  3  2  1
//      1      3GPP type = 5
//      2      3GPP Length= L
//      3 - L     UTF-8 encoded QoS profile
//
//      3GPP Type: 5
//      Length: For GGSN, L ? 37 (release 7 or higher) or L ? 33 (release 6 or release 5) or L ? 27 (release 4 or release 99) or L = 11 (release 98). For P-GW, the length varies depending on the value of QCI. See below for details.
//      QoS profile value: Text type
//      UTF-8 encoded QoS profile syntax:
//      "<Release indicator>  <release specific QoS IE UTF-8 encoding>"
//      <Release indicator> = UTF-8 encoded number (two characters) :
//       For GGSN:
//      "98" = Release 98
//      "99"= Release 99 or release 4
//      "05"= Release 5 or release 6
//      "07"= Release 7 or higher
//      For P-GW:
//      "08"= Release 8 or higher
//         <release specific QoS profile UTF-8 encoding> = UTF-8 encoded QoS profile for the release indicated by          the release indicator.
//       The UTF-8 encoding of a QoS IE is defined as follows: each octet is described by 2 UTF-8 encoded characters, defining its hexadecimal representation.
//       For GGSN:
//       The QoS profile definition is in 3GPP TS 24.008 [54].
//          The release 98 QoS profile data is 3 octets long, which then results in a 6 octets UTF-8 encoded string.
//          The release 99 and release 4 QoS profile data is 11 octets long, which results in a 22 octets UTF-8 encoded string.
//       The release 5 and release 6 QoS profile data is 14 octets long, which results in a 28 octets UTF-8 encoded string.
//       The release 7 (and higher) QoS profile data is 16 octets long, which results in a 32 octets UTF-8 encoded string.
//      For P-GW:
//       It contains the following QoS parameters associated with the EPS bearer:
//      -   QCI
//      -   ARP
//      -   GBR QoS information (UL/DL MBR, UL/DL GBR) or UL/DL APN-AMBR. In other words if the value of QCI indicates a GBR bearer, the GBR QoS information shall be present. If the value of QCI indicates a non-GBR bearer, the APN-AMBR information shall be present.
//      The encoding of the EPS bearer QoS profile parameters is specified in 3GPP TS 29.274 [81]: ARP is specified in Bearer QoS IE; QCI, UL MBR, DL MBR, UL MBR and DL MBR are specified in Flow QoS IE; UL APN-AMBR and DL APN-AMBR are specified in AMBR IE.
//      For GBR QCIs, the encoding of UTF-8 encoded QoS Profile field shall be as follows:
//
//      1-3    <Release indicator> -" = "08-" (UTF-8 encoded)
//      4-5    ARP (UTF-8 encoded)
//      6-7    QCI (UTF-8 encoded)
//      8-m    UL MBR (UTF-8 encoded)
//      (m+1)-n      DL MBR (UTF-8 encoded)
//      (n+1)-o      UL GBR (UTF-8 encoded)
//      (o+1)-p      DL GBR (UTF-8 encoded)
//
//      For non-GBR QCIs, the UL/DL MBR and UL/DL GBR fields shall not be present; UL APN-AMBR and DL APN-AMBR fields shall be encoded (in UTF-8 encoded format) respectively after the QCI field.
//
//
//
//
//      // Obtiene los parametros trafficClass y handlingPriority a partir de la cadena de entrada(proviene del AVP de la peticion) y segn la release
//      std::string analisisQoS::obtenerParamQoS(const char * GPRS_Neg_QoS_Profile)
//      {
//        TRACEAR_OPERACION (TrazarMetodoEx("analisisQoS","obtenerParamQoS",  FILE_LOCATION));
//
//        // Nos llega una cadena de caracteres
//        // Los dos primeros octetos son la release y el tercer octeto es un '-' ,
//        // el resto de octetos dependen de la release, se describen segun el caso
//        // RELEASE 99
//        char str_release[2];
//        strncpy( str_release,GPRS_Neg_QoS_Profile,2);
//        //qosProfile ser el resto de la cadena ( a partir del guin)
//        char qosProfile[MAX_CADENA_QOS+1]="";
//        char cadena[MAX_CADENA_QOS+1]="";
//        // convierte el formato UTF8 a string,
//        strncpy(cadena,GPRS_Neg_QoS_Profile,MAX_CADENA_QOS+1);
//
//        // RELEASE_99;
//
//          if (cadena [0] =='9' && cadena [1]=='9')
//          {
//      10.5.6.5  Quality of service
//      The purpose of the quality of service information element is to specify the QoS parameters for a PDP context.
//      The QoS IE is defined to allow backward compatibility to earlier version of Session Management Protocol.
//      The quality of service is a type 4 information element with a minimum length of 14 octets and a maximum length of 18 octets. The QoS requested by the MS shall be encoded both in the QoS attributes specified in octets 3-5 and in the QoS attributes specified in octets 6-14.
//      In the MS to network direction and in the network to MS direction the following applies:
//      -   Octets 15-18 are optional. If octet 15 is included, then octet 16 shall also be included, and octets 17 and 18 may be included.
//      -   If octet 17 is included, then octet 18 shall also be included.
//      -   A QoS IE received without octets 6-18, without octets 14-18, without octets 15-18, or without octets 17-18 shall be accepted by the receiving entity.
//      NOTE:  This behavior is required for interworking with entities supporting an earlier version of the protocol, or when the Maximum bit rate for downlink or for downlink and uplink is negotiated to a value lower than 8700 kbps.
//      The quality of service information element is coded as shown in figure 10.5.138/3GPP TS 24.008 and table 10.5.156/3GPP TS 24.008.
//
//      Figure 10.5.138/3GPP TS 24.008: Quality of service information element
//            //          8           7           6           5           4           3           2           1
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                   Quality of service IEI                                      +  Octet 1
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                               Length of quality of service IE                                 +  Octet 2
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    |         spare(00)     +          Delay class              +     Reliability class             +  Octet 3
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +              Peak throughput                  +   spare   +    Precedence class               +  Octet 4
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +             spare(000)            +                  Mean throughput                          +  Octet 5
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +          Traffic class            +     Delivery order    +     Delivery of erroneus SDU      +  Octet 6 <---
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                       Maximum SDU size                                        +  Octet 7
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                  Maximum bit rate for uplink                                  +  Octet 8
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                 Maximum bit rate for downlink                                 +  Octet 9
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                  Residual BER                 +               SDU error ratio                 +  Octet 10
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                      Transfer delay                                   +    Traffic Handling   +  Octet 11 <--
//            //    +                                                                       +        priority       +
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                 Guaranteed bit rate for uplink                                +  Octet 12
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                                Guaranteed bit rate for downlink                               +  Octet 13
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +             spare(000)            +Signalling +             Source Statistics Descriptor      +  Octet 14
//            //    +                                   +Indication +                                               +
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                             Maximum bit rate for downlink (extended)                          +  Octet 15
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                            Guaranteed bit rate for downlink (extended)                        +  Octet 16
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                              Maximum bit rate for uplink (extended)                           +  Octet 17
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +                             Guaranteed bit rate for uplink (extended)                          +  Octet 18
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//
//      Reliability class, octet 3 (see 3GPP TS 23.107 [81])
//      Bits
//      3 2 1
//      In MS to network direction:
//      0 0 0  Subscribed reliability class
//      In network to MS direction:
//      0 0 0  Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 1  Unused. If received, it shall be interpreted as '010' (Note)
//      0 1 0  Unacknowledged GTP; Acknowledged LLC and RLC, Protected data
//      0 1 1  Unacknowledged GTP and LLC; Acknowledged RLC, Protected data
//      1 0 0  Unacknowledged GTP, LLC, and RLC, Protected data
//      1 0 1  Unacknowledged GTP, LLC, and RLC, Unprotected data
//      1 1 1  Reserved
//
//      All other values are interpreted as Unacknowledged GTP and LLC; Acknowledged RLC, Protected data in this version of the protocol.
//
//      If network supports EPS, then it should not assign Reliability class value 010.
//
//      NOTE: this value was allocated in earlier versions of the protocol.
//
//      Delay class, octet 3 (see 3GPP TS 22.060 [73] and 3GPP TS 23.107 [81])
//      Bits
//      6 5 4
//      In MS to network direction:
//      0 0 0  Subscribed delay class
//      In network to MS direction:
//      0 0 0  Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 1  Delay class 1
//      0 1 0  Delay class 2
//      0 1 1  Delay class 3
//      1 0 0  Delay class 4 (best effort)
//      1 1 1  Reserved
//
//
//
//      All other values are interpreted as Delay class 4 (best effort) in this version
//      of the protocol.
//      Bit 7 and 8 of octet 3 are spare and shall be coded all 0.
//      Precedence class, octet 4 (see 3GPP TS 23.107 [81])
//      Bits
//      3 2 1
//      In MS to network direction:
//      0 0 0  Subscribed precedence
//      In network to MS direction:
//      0 0 0  Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 1  High priority
//      0 1 0  Normal priority
//      0 1 1  Low priority
//      1 1 1  Reserved
//
//
//      All other values are interpreted as Normal priority in this version of the protocol.
//
//      Bit 4 of octet 4 is spare and shall be coded as 0.
//
//      Peak throughput, octet 4 (see 3GPP TS 23.107 [81])
//      This field is the binary representation of the Peak Throughput Class (1 to 9). The corresponding peak throughput to each peak throughput class is indicated.
//      Bits
//      8 7 6 5
//      In MS to network direction:
//      0 0 0 0      Subscribed peak throughput
//      In network to MS direction:
//      0 0 0 0      Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 0 1      Up to 1 000 octet/s
//      0 0 1 0      Up to 2 000 octet/s
//      0 0 1 1      Up to 4 000 octet/s
//      0 1 0 0      Up to 8 000 octet/s
//      0 1 0 1      Up to 16 000 octet/s
//      0 1 1 0      Up to 32 000 octet/s
//      0 1 1 1      Up to 64 000 octet/s
//      1 0 0 0      Up to 128 000 octet/s
//      1 0 0 1      Up to 256 000 octet/s
//      1 1 1 1      Reserved
//
//      All other values are interpreted as Up to 1 000 octet/s in this
//      version of the protocol.
//      Mean throughput, octet 5 (see 3GPP TS 23.107 [81])
//      This field is the binary representation of the Mean Throughput Class (1 to 18; mean throughput class 30 is reserved and 31 is best effort). The corresponding mean throughput to each mean throughput class is indicated.
//      Bits
//      5 4 3 2 1
//
//
//
//      In MS to network direction:
//      0 0 0 0 0    Subscribed mean throughput
//      In network to MS direction:
//      0 0 0 0 0    Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 0 0 1    100 octet/h
//      0 0 0 1 0    200 octet/h
//      0 0 0 1 1    500 octet/h
//      0 0 1 0 0    1 000 octet/h
//      0 0 1 0 1    2 000 octet/h
//      0 0 1 1 0    5 000 octet/h
//      0 0 1 1 1    10 000 octet/h
//      0 1 0 0 0    20 000 octet/h
//      0 1 0 0 1    50 000 octet/h
//      0 1 0 1 0    100 000 octet/h
//      0 1 0 1 1    200 000 octet/h
//      0 1 1 0 0    500 000 octet/h
//      0 1 1 0 1    1 000 000 octet/h
//      0 1 1 1 0    2 000 000 octet/h
//      0 1 1 1 1    5 000 000 octet/h
//      1 0 0 0 0    10 000 000 octet/h
//      1 0 0 0 1    20 000 000 octet/h
//      1 0 0 1 0    50 000 000 octet/h
//      1 1 1 1 0    Reserved
//      1 1 1 1 1    Best effort
//      The value Best effort indicates that throughput shall be made available to the MS on a per need and availability basis.
//      All other values are interpreted as Best effort in this
//      version of the protocol.
//
//      Bits 8 to 6 of octet 5 are spare and shall be coded all 0.
//
//
//      Delivery of erroneous SDUs, octet 6 (see 3GPP TS 23.107 [81])
//      Bits
//      3 2 1
//      In MS to network direction:
//      0 0 0     Subscribed delivery of erroneous SDUs
//      In network to MS direction:
//      0 0 0     Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 1     No detect ('-')
//      0 1 0     Erroneous SDUs are delivered ('yes')
//      0 1 1     Erroneous SDUs are not delivered ('no')
//      1 1 1     Reserved
//
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of this protocol.
//
//      The MS shall consider all other values as reserved.
//
//      Delivery order, octet 6 (see 3GPP TS 23.107 [81])
//      Bits
//      5 4 3
//      In MS to network direction:
//      0 0    Subscribed delivery order
//      In network to MS direction:
//      0 0    Reserved
//      In MS to network direction and in network to MS direction:
//      0 1    With delivery order ('yes')
//      1 0    Without delivery order ('no')
//      1 1    Reserved
//
//
//
//
//      Traffic class, octet 6 (see 3GPP TS 23.107 [81])
//      Bits
//      8 7 6
//      In MS to network direction:
//      0 0 0     Subscribed traffic class
//      In network to MS direction:
//      0 0 0     Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 1     Conversational class
//      0 1 0     Streaming class
//      0 1 1     Interactive class
//      1 0 0     Background class
//      1 1 1     Reserved
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of this protocol.
//
//
//      The MS shall consider all other values as reserved.
//
//      Maximum SDU size, octet 7 (see 3GPP TS 23.107 [81])
//      In MS to network direction:
//      0 0 0 0 0 0 0 0    Subscribed maximum SDU size
//      1 1 1 1 1 1 1 1    Reserved
//      In network to MS direction:
//      0 0 0 0 0 0 0 0    Reserved
//      1 1 1 1 1 1 1 1    Reserved
//      In MS to network direction and in network to MS direction:
//
//      For values in the range 00000001 to 10010110 the Maximum SDU size value is binary coded in 8 bits, using a granularity of 10 octets, giving a range of values from 10 octets to 1500 octets.
//      Values above 10010110 are as below:
//      1 0 0 1 0 1 1 1    1502 octets
//      1 0 0 1 1 0 0 0    1510 octets
//      1 0 0 1 1 0 0 1    1520 octets
//
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of this protocol.
//
//      The MS shall consider all other values as reserved.
//
//
//      Maximum bit rate for uplink, octet 8
//      Bits
//      8 7 6 5 4 3 2 1
//      In MS to network direction:
//      0 0 0 0 0 0 0 0 Subscribed maximum bit rate for uplink
//      In network to MS direction:
//      0 0 0 0 0 0 0 0 Reserved
//      In MS to network direction and in network to MS direction:
//      0 0 0 0 0 0 0 1    The maximum bit rate is binary coded in 8 bits, using a granularity of 1 kbps
//      0 0 1 1 1 1 1 1 giving a range of values from 1 kbps to 63 kbps in 1 kbps increments.
//
//      0 1 0 0 0 0 0 0    The maximum bit rate is 64 kbps + ((the binary coded value in 8 bits 01000000) * 8 kbps)
//      0 1 1 1 1 1 1 1 giving a range of values from 64 kbps to 568 kbps in 8 kbps increments.
//
//      1 0 0 0 0 0 0 0    The maximum bit rate is 576 kbps + ((the binary coded value in 8 bits 10000000) * 64 kbps)
//      1 1 1 1 1 1 1 0 giving a range of values from 576 kbps to 8640 kbps in 64 kbps increments.
//
//      1 1 1 1 1 1 1 1 0kbps
//
//      If the sending entity wants to indicate a Maximum bit rate for uplink higher than 8640 kbps, it shall set octet 8 to 11111110, i.e. 8640 kbps, and shall encode the value for the Maximum bit rate in octet 17.
//
//
//      Maximum bit rate for downlink, octet 9 (see 3GPP TS 23.107 [81])
//
//      Coding is identical to that of Maximum bit rate for uplink.
//
//      If the sending entity wants to indicate a Maximum bit rate for downlink higher than 8640 kbps, it shall set octet 9 to 11111110, i.e. 8640 kbps, and shall encode the value for the Maximum bit rate in octet 15.
//
//      In this version of the protocol, for messages specified in the present document, the sending entity shall not request 0 kbps for both the Maximum bitrate for downlink and the Maximum bitrate for uplink at the same time. Any entity receiving a request for 0 kbps in both the Maximum bitrate for downlink and the Maximum bitrate for uplink shall consider that as a syntactical error (see clause 8).
//
//
//
//      Residual Bit Error Rate (BER), octet 10 (see 3GPP TS 23.107 [81])
//      Bits
//      8 7 6 5
//      In MS to network direction:
//      0 0 0 0      Subscribed residual BER
//      In network to MS direction:
//      0 0 0 0      Reserved
//      In MS to network direction and in network to MS direction:
//      The Residual BER value consists of 4 bits. The range is from 5*10-2 to 6*10-8.
//      0 0 0 1      5*10-2
//      0 0 1 0      1*10-2
//      0 0 1 1      5*10-3
//      0 1 0 0      4*10-3
//      0 1 0 1      1*10-3
//      0 1 1 0      1*10-4
//      0 1 1 1      1*10-5
//      1 0 0 0      1*10-6
//      1 0 0 1      6*10-8
//      1 1 1 1      Reserved
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall consider all other values as reserved.
//
//      SDU error ratio, octet 10 (see 3GPP TS 23.107 [81])
//      Bits
//      4 3 2 1
//      In MS to network direction:
//      0 0 0 0      Subscribed SDU error ratio
//      In network to MS direction:
//      0 0 0 0      Reserved
//      In MS to network direction and in network to MS direction:
//      The SDU error ratio value consists of 4 bits. The range is is from 1*10-1 to 1*10-6.
//      0 0 0 1      1*10-2
//      0 0 1 0      7*10-3
//      0 0 1 1      1*10-3
//      0 1 0 0      1*10-4
//      0 1 0 1      1*10-5
//      0 1 1 0      1*10-6
//      0 1 1 1      1*10-1
//      1 1 1 1      Reserved
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall consider all other values as reserved.
//
//      Traffic handling priority, octet 11 (see 3GPP TS 23.107 [81])
//      Bits
//      2 1
//      In MS to network direction:
//      0 0    Subscribed traffic handling priority
//      In network to MS direction:
//      0 0    Reserved
//      In MS to network direction and in network to MS direction:
//      0 1    Priority level 1
//      1 0    Priority level 2
//      1 1    Priority level 3
//
//      The Traffic handling priority value is ignored if the Traffic Class is Conversational class, Streaming class or Background class.
//
//      Transfer delay, octet 11 (See 3GPP TS 23.107 [81])
//      Bits
//      8 7 6 5 4 3
//
//
//
//      In MS to network direction:
//      0 0 0 0 0 0     Subscribed transfer delay
//      In network to MS direction:
//      0 0 0 0 0 0     Reserved
//      In MS to network direction and in network to MS direction:
//
//
//      0 0 0 0 0 1  The Transfer delay is binary coded in 6 bits, using a granularity of 10 ms
//      0 0 1 1 1 1     giving a range of values from 10 ms to 150 ms in 10 ms increments
//
//      0 1 0 0 0 0  The transfer delay is 200 ms + ((the binary coded value in 6 bits  010000) * 50 ms)
//      0 1 1 1 1 1     giving a range of values from 200 ms to 950 ms in 50ms increments
//
//      1 0 0 0 0 0  The transfer delay is 1000 ms + ((the binary coded value in 6 bits  100000) * 100 ms)
//      1 1 1 1 1 0     giving a range of values from 1000 ms to 4000 ms in 100ms increments
//
//      1 1 1 1 1 1     Reserved
//
//      The Transfer delay value is ignored if the Traffic Class is Interactive class or Background class.
//
//      Guaranteed bit rate for uplink, octet 12 (See 3GPP TS 23.107 [81])
//
//      Coding is identical to that of Maximum bit rate for uplink.
//
//      If the sending entity wants to indicate a Guaranteed bit rate for uplink higher than 8640 kbps, it shall set octet 12 to 11111110, i.e. 8640 kbps, and shall encode the value for the Guaranteed bit rate in octet 18.
//
//      The Guaranteed bit rate for uplink value is ignored if the Traffic Class is Interactive class or Background class, or Maximum bit rate for uplink is set to 0 kbps.
//
//      Guaranteed bit rate for downlink, octet 13(See 3GPP TS 23.107 [81])
//
//      Coding is identical to that of Maximum bit rate for uplink.
//
//      If the sending entity wants to indicate a Guaranteed bit rate for downlink higher than 8640 kbps, it shall set octet 13 to 11111110, i.e. 8640 kbps, and shall encode the value for the Guaranteed bit rate in octet 16.
//
//      The Guaranteed bit rate for downlink value is ignored if the Traffic Class is Interactive class or Background class, or Maximum bit rate for downlink is set to 0 kbps.
//
//      Source Statistics Descriptor, octet 14 (see 3GPP TS 23.107 [81])
//      Bits
//      4 3 2 1
//      In MS to network direction
//      0 0 0 0   unknown
//      0 0 0 1      speech
//
//      The network shall consider all other values as unknown.
//
//      In network to MS direction
//      Bits 4 to 1 of octet 14 are spare and shall be coded all 0.
//
//      The Source Statistics Descriptor value is ignored if the Traffic Class is Interactive class or Background class.
//
//      Signalling Indication, octet 14 (see 3GPP TS 23.107 [81])
//      Bit
//      5
//      In MS to network direction and in network to MS direction:
//      0      Not optimised for signalling traffic
//      1      Optimised for signalling traffic
//
//      If set to '1' the QoS of the PDP context is optimised for signalling
//
//      The Signalling Indication value is ignored if the Traffic Class is Conversational class, Streaming class or Background class.
//
//      Bits 8 to 6 of octet 14 are spare and shall be coded all 0.
//
//      Maximum bit rate for downlink (extended), octet 15
//      Bits
//      8 7 6 5 4 3 2 1
//      In MS to network direction and in network to MS direction:
//      0 0 0 0 0 0 0 0 Use the value indicated by the Maximum bit rate for downlink in octet 9.
//
//                   For all other values: Ignore the value indicated by the Maximum bit rate for downlink in octet 9
//                   and use the following value:
//      0 0 0 0 0 0 0 1 The maximum bit rate is 8600 kbps + ((the binary coded value in 8 bits) * 100 kbps),
//      0 1 0 0 1 0 1 0 giving a range of values from 8700 kbps
//      to 16000 kbps in 100 kbps increments.
//
//      0 1 0 0 1 0 1 1 The maximum bit rate is 16 Mbps + ((the binary coded value in 8 bits - 01001010) * 1 Mbps),
//      1 0 1 1 1 0 1 0 giving a range of values from 17 Mbps to 128 Mbps in 1 Mbps increments.
//
//      1 0 1 1 1 0 1 1 The maximum bit rate is 128 Mbps + ((the binary coded value in 8 bits - 10111010) * 2 Mbps),
//      1 1 1 1 1 0 1 0 giving a range of values from 130 Mbps to 256 Mbps in 2 Mbps increments.
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall map all other values not explicitly defined onto the maximum value defined in this version of the protocol.
//
//      Guaranteed bit rate for downlink (extended), octet 16
//      Bits
//      8 7 6 5 4 3 2 1
//      In MS to network direction and in network to MS direction:
//      0 0 0 0 0 0 0 0 Use the value indicated by the Guaranteed bit rate for downlink in octet 13.
//
//                   For all other values: Ignore the value indicated by the Guaranteed bit rate for downlink in octet 9
//                   and use the following value:
//      0 0 0 0 0 0 0 1 The guaranteed bit rate is 8600 kbps + ((the binary coded value in 8 bits) * 100 kbps),
//      0 1 0 0 1 0 1 0 giving a range of values from 8700 kbps to 16000 kbps in 100 kbps increments.
//
//      0 1 0 0 1 0 1 1 The guaranteed bit rate is 16 Mbps + ((the binary coded value in 8 bits - 01001010) * 1 Mbps),
//      1 0 1 1 1 0 1 0 giving a range of values from 17 Mbps to 128 Mbps in 1 Mbps increments.
//
//      1 0 1 1 1 0 1 1 The guaranteed bit rate is 128 Mbps + ((the binary coded value in 8 bits - 10111010) * 2 Mbps),
//      1 1 1 1 1 0 1 0 giving a range of values from 130 Mbps to 256 Mbps in 2 Mbps increments.
//
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall map all other values not explicitly defined onto the maximum value defined in this version of the protocol.
//
//      Maximum bit rate for uplink (extended), octet 17
//
//      This field is an extension of the Maximum bit rate for uplink in octet 8. The coding is identical to that of the Maximum bit rate for downlink (extended).
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall map all other values not explicitly defined onto the maximum value defined in this version of the protocol.
//
//      Guaranteed bit rate for uplink (extended), octet 18
//
//      This field is an extension of the Guaranteed bit rate for uplink in octet 12. The coding is identical to that of the Guaranteed bit rate for downlink (extended).
//
//      The network shall map all other values not explicitly defined onto one of the values defined in this version of the protocol. The network shall return a negotiated value which is explicitly defined in this version of the protocol.
//
//      The MS shall map all other values not explicitly defined onto the maximum value defined in this version of the protocol.
//
//
//            //trafficClass: Octeto 4 , //strtok(cadena,"-");
//            //--------------------------------------------------------------------
//            a_release = 99;
//            char octeto4[3]= "\0";
//            // CADENA: caracteres 0,1, release, caracter 2 " - ", car. 3,4: octeto1, 4,5: octeto2....12,13: octeto11
//            //octetoi  se obtiene de cadena[2i+1] , cadena[2i+2], dos caracteres en hexadecimal
//
//            octeto4[0]=cadena[9];
//            octeto4[1]=cadena[10];
//            octeto4[2]='\0';
//            long i_octeto4;
//            // pasamos de hexadecimal a decimal
//            sscanf(octeto4,"%x",&i_octeto4);
//            a_trafficClass = (i_octeto4 & 0xE0)>>5;
//            TRACEAR_DEPURACION
//       (
//        trazaDebInt("Parametro TrafficClass recibido", a_trafficClass);
//        );
//            //trafficHandlingPriority : octeto 9, bits 1,2
//            //------------------------------------------------------------------
//            char octeto9[3]= "\0";
//            octeto9[0]=cadena[19];
//            octeto9[1]=cadena[20];
//            octeto9[2]='\0';
//            long i_octeto9;
//
//            // lo paso a decimal
//            sscanf(octeto9,"%x",&i_octeto9);
//            a_handlingPriority = (i_octeto9 & 0x03);
//            TRACEAR_DEPURACION
//       (
//        trazaDebInt("Parametro TrafficHandlingPriority recibido",a_handlingPriority);
//        );
//
//              //Maximum bit rate for uplink : octeto 6 completo
//            //------------------------------------------------------------------
//
//             char octeto6[3]= "\0";
//
//             octeto6[0]=cadena[13];
//             octeto6[1]=cadena[14];
//             octeto6[2]='\0';
//
//             a_maximumUplink = decodificaBitRate(octeto6);
//
//             trazaDebInt("Parametro Maximum bit rate for uplink recibido(en Kbps)",a_maximumUplink);
//
//              //Maximum bit rate for downlink : octeto 7 completo
//            //------------------------------------------------------------------
//
//             char octeto7[3]= "\0";
//
//             octeto7[0]=cadena[15];
//             octeto7[1]=cadena[16];
//             octeto7[2]='\0';
//
//             a_maximumDownlink = decodificaBitRate(octeto7);
//
//             trazaDebInt("Parametro Maximum bit rate for downlink recibido(en Kbps)",a_maximumDownlink);
//
//
//
//            // qosProfile es la cadena a quitando la release (tres primeros caracteres) 98-qosProfile o 99-qosProfile
//            //-----------------
//            strncpy(qosProfile,cadena+3,MAX_CADENA_QOS-2);
//            TRACEAR_DEPURACION
//       (
//        trazaDebStr("Parametro qosProfile recibido",qosProfile);
//        );
//
//            return std::string(qosProfile);
//          }
//
//        //RELEASE_98;
//        else if (cadena [0] =='9' && cadena [1]=='8')
//          {
//
//            //           8          7            6            5           4           3           2             1
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    |         spare        +          Delay class          +     Reliability class         +  Octeto 1
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//            //    +              Peak throughput           + spare +    Precedence class        +  Octeto 2
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - -       +
//            //    +        spare(000)            +                  Mean throughput                    +  Octeto 3
//            //    + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - + - - - - - +
//
//            //Delay Class . Octeto 1 , bits 6,5,4
//            //------------------
//            a_release = 98;
//            char octeto1[3]= "\0";
//            //octetoi  es cadena[2i+1],cadena[2i+2] en hexadecimal
//            octeto1[0]=cadena[3];
//            octeto1[1]=cadena[4];
//            octeto1[2]='\0';
//            // lo paso a decimal
//            long i_octeto1=0;
//            sscanf(octeto1,"%x",&i_octeto1);
//            int delayClass = (i_octeto1 & 0x38)>>3;
//            TRACEAR_DEPURACION
//       (
//        trazaDebInt("Parametro DelayClass recibido",delayClass);
//        );
//            a_trafficClass = delayClass; // Devolvemos el valor en la var trafficClass
//            //Precedence Class . Octeto 2, bits 1,2,3
//            //--------------------------
//            char octeto2[3]= "\0";
//            octeto2[0]=cadena[5];
//            octeto2[1]=cadena[6];
//            octeto2[2]='\0';
//            //en decimal
//            long i_octeto2;
//            sscanf(octeto2,"%x",&i_octeto2);
//
//            int precedenceClass = i_octeto2 & 0x07;
//
//             TRACEAR_DEPURACION
//       (
//        trazaDebInt("Parametro precedenceClass recibido: ",precedenceClass);
//        );
//
//            a_handlingPriority =  precedenceClass; // Devolvemos el valor en la var trafficHandPriority
//            // qosProfile
//            strcpy(qosProfile, cadena+3);// qosProfile, es la cadena quitandole la release, ->"98-qosProfile"
//            TRACEAR_DEPURACION
//       (
//        trazaDebStr( "Parametro qosProfile recibido",qosProfile);
//        );
//
//            return std::string(qosProfile);
//          }//fin release98
//        else
//          {
//             TRACEAR_ERROR
//       (
//        trazaErrStr(TR("El parametro Release no es valido ( debe esr 98 o 99)"),str_release);
//        );
//             return std::string("");
//          }
//      }
//
//      int analisisQoS::decodificaBitRate(char * dato)
//      {
//       int cabecera=0;
//       int salida=0;
//       char octeto[3]= "\0";
//       long i_octeto;
//
//       strcpy(octeto,dato);
//
//           /*Codificacion de los bits de maxima tasa de bit
//          Maximum bit rate for uplink, octet 8
//          Bits
//          8 7 6 5 4 3 2 1
//          In MS to network direction:
//          0 0 0 0 0 0 0 0   Subscribed maximum bit rate for uplink
//          In network to MS direction:
//          0 0 0 0 0 0 0 0   Reserved
//          In MS to network direction and in network to MS direction:
//          0 0 0 0 0 0 0 1   The maximum bit rate is binary coded in 8 bits, using a granularity of 1 kbps
//          0 0 1 1 1 1 1 1   giving a range of values from 1 kbps to 63 kbps in 1 kbps increments.
//
//          0 1 0 0 0 0 0 0   The maximum bit rate is 64 kbps +  ((the binary coded value in 8 bits 01000000) * 8 kbps)
//          0 1 1 1 1 1 1 1   giving a range of values from 64 kbps to 568 kbps in 8 kbps increments.
//
//          1 0 0 0 0 0 0 0   The maximum bit rate is 576 kbps + ((the binary coded value in 8 bits 10000000) * 64 kbps)
//          1 1 1 1 1 1 1 0   giving a range of values from 576 kbps to 8640 kbps in 64 kbps increments.
//
//          1 1 1 1 1 1 1 1   0kbps
//
//
//          Maximum bit rate for downlink, octet 9 (see 3GPPTS 23.107)
//
//          Coding is identical to that of Maximum bit rate for uplink.
//
//         */
//
//       // lo paso a decimal
//       sscanf(octeto,"%x",&i_octeto);
//       cabecera = (i_octeto & 0xC0)>>6;
//
//       trazaDebStr("QoS, octeto: ", octeto);
//       trazaDebInt("QoS, i_octeto: ", i_octeto);
//       trazaDebInt("QoS, cabecera: ", cabecera);
//
//       //Casos especiales: El caso 00000000 se considera 0, por no poder definir ninguno de los dos
//       // casos de la especificacion(mirar codificacion arriba)
//
//       if(i_octeto != 255 && i_octeto != 0)
//       {
//          switch(cabecera)
//          {
//             case 0:
//                trazaDeb("Bitrate con granuralidad de 1 Kbps");
//                salida = (i_octeto & 0x3F);
//             break;
//
//             case 1:
//                trazaDeb("Bitrate con granuralidad de 8 Kbps");
//                salida = 64 + (i_octeto & 0x3F) * 8;
//             break;
//
//             case 2:
//             case 3:
//                trazaDeb("Bitrate con granuralidad de 64 Kbps");
//                salida = 576 + (i_octeto & 0x7E) * 64;
//             break;
//
//             default:
//                trazaErr("Valor de bitrate no contemplado. Revisar codificacion");
//                salida = 0;
//             break;
//          }
//       }
//       else
//       {
//          trazaDeb("Bitrate vale 255 o 0, valores no contemplados");
//          salida = 0;
//       }
//       return salida;
//      }

