// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/helpers/dcca/defines.hpp>
#include <anna/diameter/helpers/dcca/functions.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/config/defines.hpp> // general types, decoding helpers (DECODE[2/3/4]BYTES_INDX_VALUETYPE), etc.
#include <anna/core/functions.hpp>

#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>

// STL
#include <string>

using namespace anna;
using namespace anna::diameter::codec;
using namespace anna::diameter::helpers::dcca;



//static
const char anna::diameter::helpers::dcca::ChargingContextAndDomainSuffix::Data[] = "32251@3gpp.org";
const char anna::diameter::helpers::dcca::ChargingContextAndDomainSuffix::Voice[] = "OCS-CS@telefonica.com";
const char anna::diameter::helpers::dcca::ChargingContextAndDomainSuffix::Content[] = "OCS-Generic-Services@telefonica.com";
const char anna::diameter::helpers::dcca::ChargingContextAndDomainSuffix::SMS[] = "32274@3gpp.org";
const char anna::diameter::helpers::dcca::ChargingContextAndDomainSuffix::MMS[] = "32270@3gpp.org";


// getters
std::string anna::diameter::helpers::dcca::functions::getSubscriptionIdData(const anna::DataBlock & db, int subscriptionIdType) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  // Message datablock:
  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;

  // Auxiliar avp decoding:
  AvpId _id;
  char _flags;
  int _length;
  int type;
  std::string _dataG /* grouped avp */, _data1 /* first avp within grouped */, _data2 /* second avp within grouped */;

  const char * subscriptionIdPtr;
  int pos = 1; // first subscriber to find
  while((subscriptionIdPtr = diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__Subscription_Id, pos))) {

    // Decode data for the grouped avp found:
    diameter::codec::functions::decodeAVP(subscriptionIdPtr, _id, _flags, _length, _dataG);
    // Data (_dataG) consists in two mandatory Avps in ANY order:
    //  - Subscription-Id-Type (Enumerated derived from Integer32)
    //  - Subscription-Id-Data (UTF8String)

    // First AVP within _dataG:
    diameter::codec::functions::decodeAVP(_dataG.c_str(), _id, _flags, _length, _data1);

    if (_id == diameter::helpers::dcca::AVPID__Subscription_Id_Type) {
      type = DECODE4BYTES_INDX_VALUETYPE(_data1, 0, int);
      if (type == subscriptionIdType) {
        diameter::codec::functions::decodeAVP(_dataG.c_str() + 12, _id, _flags, _length, _data2);
        if (_id == diameter::helpers::dcca::AVPID__Subscription_Id_Data) return _data2;
      }
    }
    else if (_id == diameter::helpers::dcca::AVPID__Subscription_Id_Data) {
      diameter::codec::functions::decodeAVP(_dataG.c_str() + 20, _id, _flags, _length, _data2);
      if (_id == diameter::helpers::dcca::AVPID__Subscription_Id_Type) {
        type = DECODE4BYTES_INDX_VALUETYPE(_data2, 0, int);
        if (type == subscriptionIdType) return _data1;
      }
    }

    pos++;
  }


  // Nothing retrieved:
  return "";
}


std::string anna::diameter::helpers::dcca::functions::getServiceContextId(const anna::DataBlock & db, ChargingContext::_v &chargingContext) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  //anna::DataBlock avpsDB(db.getData() + Message::HeaderLength, db.getSize() - Message::HeaderLength);
  //const char * serviceContextIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, AVPID__Service_Context_Id);

  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;
  const char * serviceContextIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__Service_Context_Id);

  if(serviceContextIdPtr == NULL)
    throw anna::RuntimeException("Service-Context-Id AVP not found in DataBlock provided", ANNA_FILE_LOCATION);

  // Decoded avp information:
  AvpId _id;
  char _flags;
  int _length;
  std::string result;
  anna::diameter::codec::functions::decodeAVP(serviceContextIdPtr, _id, _flags, _length, result /* data-part */);
  // Charging context detection:
  chargingContext = ChargingContext::Unknown;

  if(anna::functions::endsWith(result, ChargingContextAndDomainSuffix::Data))
    chargingContext = ChargingContext::Data;
  else if(anna::functions::endsWith(result, ChargingContextAndDomainSuffix::Voice))
    chargingContext = ChargingContext::Voice;
  else if(anna::functions::endsWith(result, ChargingContextAndDomainSuffix::Content))
    chargingContext = ChargingContext::Content;
  else if(anna::functions::endsWith(result, ChargingContextAndDomainSuffix::SMS))
    chargingContext = ChargingContext::SMS;
  else if(anna::functions::endsWith(result, ChargingContextAndDomainSuffix::MMS))
    chargingContext = ChargingContext::MMS;

  // ...
  // future kind of traffic
  return result;
}

