// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/helpers/ericsson/defines.hpp>
#include <anna/diameter/helpers/ericsson/functions.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/config/defines.hpp> // general types, decoding helpers (DECODE[2/3/4]BYTES_INDX_VALUETYPE), etc.

#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>

// STL
#include <string>

using namespace anna;
using namespace anna::diameter::codec;
using namespace anna::diameter::helpers::ericsson;


// getters
std::string anna::diameter::helpers::ericsson::functions::getSCAPSubscriptionIdData(const anna::DataBlock & db, int scapSubscriptionIdType) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  //anna::DataBlock avpsDB(db.getData() + Message::HeaderLength, db.getSize() - Message::HeaderLength);
  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;
  std::string result = "";
  bool found = false;
  int pos = 1; // first avp
  const char * subscriptionIdPtr;
  const char * subscriptionIdDataPtr;
  int type;
  // Decoded avp information:
  AvpId _id;
  char _flags;
  int _length;
  std::string _dataG /* grouped */, _data;

  while(!found) {
    //subscriptionIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, AVPID__SCAP_Subscription_Id, pos);
    subscriptionIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__SCAP_Subscription_Id, pos);

    if(!subscriptionIdPtr) return result;

    // Look up type:
    anna::diameter::codec::functions::decodeAVP(subscriptionIdPtr, _id, _flags, _length, _dataG);
    // Data is Fixed SCAP-Subscription-Id-Type (Enumerated derived from Integer32) and then Fixed SCAP-Subscription-Id-Data (UTF8String):
    // No need to find SCAP-Subscription-Id-Type, it's always the first:
    anna::diameter::codec::functions::decodeAVP(_dataG.c_str(), _id, _flags, _length, _data);
    // Enumerated
    int type = DECODE4BYTES_INDX_VALUETYPE(_data, 0, int);
    found = (type == scapSubscriptionIdType);
    pos++;
  }

  // No need to find SCAP-Subscription-Id-Data within _dataG, it's always the second, and the first takes always 4 words (has vendorID):
  subscriptionIdDataPtr = _dataG.c_str() + 16;
  anna::diameter::codec::functions::decodeAVP(subscriptionIdDataPtr, _id, _flags, _length, result);
  return result;
}


