// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/DataBlock.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>

#include <anna/comm/ClientSocket.hpp>

#include <anna/http/Handler.hpp>
#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Transport.hpp>

using namespace std;
using namespace anna;

http::Handler::~Handler() {
  delete a_response;
}

//------------------------------------------------------------------------------------------------
// (1) Si ha llegado hasta aqui deberia ser porque el mensaje recibido es del protocolo HTTP.
//------------------------------------------------------------------------------------------------
void http::Handler::apply(comm::ClientSocket& clientSocket, const comm::Message& message)
throw(RuntimeException) {
  if(clientSocket.support(http::Transport::className()) == false) {
    LOGWARNING(
      string msg("anna::http::Handler::apply | Incoming ClientSocket has no support for HTTP | Message: ");
      msg += functions::asString(message.getBody());
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  const http::Message& httpMessage = static_cast <const http::Message&>(message);               // (1)

  LOGDEBUG(
    string msg("apply | ");
    msg += asString();
    msg += " | Message: ";
    msg += httpMessage.asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(httpMessage.getType() == Message::Type::Request) {
    try {
      evRequest(clientSocket, static_cast <const Request&>(httpMessage));
    } catch(RuntimeException& ex) {
      ex.trace();
      Response* response = allocateResponse();
      response->setStatusCode(400);
      response->setReasonPhrase(ex.asString());
      clientSocket.send(*response);
    }
  } else {
    try {
      evResponse(clientSocket, static_cast <const Response&>(httpMessage));
    } catch(RuntimeException& ex) {
      ex.trace();
    }
  }
}

http::Response* http::Handler::allocateResponse()
throw() {
  return (a_response == NULL) ? (a_response = new Response) : a_response;
}


