// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/ldap/Response.hpp>
#include <anna/ldap/Attribute.hpp>
#include <anna/ldap/TimerManager.hpp>

using namespace std;
using namespace anna;

ldap::Response::response_pool ldap::Response::st_responses;

//----------------------------------------------------------------------------------------
// Se invocan desde ldap::Session
//----------------------------------------------------------------------------------------
ldap::Response* ldap::Response::instance(const ClassCode::_v classCode, const IdMessage idMessage)
throw(RuntimeException) {
  ldap::Response* result = st_responses.create();
  result->a_classCode = classCode;
  result->a_idMessage = idMessage;
  result->clear();
  LOGDEBUG(
    string msg("anna::ldap::Response::instance | ");
    msg += result->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  return result;
}

void ldap::Response::release(ldap::Response* response)
throw() {
  try {
    st_responses.release(response);
  } catch(Exception& ex) {
    ex.trace();
  }
}

ldap::Response::Response() :
  a_classCode(ClassCode::Undefined),
  a_idMessage(0),
  a_session(NULL),
  a_timer(NULL),
  a_request(NULL) {
}

void ldap::Response::clear()
throw() {
  a_session = NULL;
  a_resultCode.clear();
  a_name.clear();

  for(attribute_iterator ii = attribute_begin(), maxii = attribute_end(); ii != maxii; ii ++)
    attribute(ii)->clear();

  a_attributes.clear();
  a_referrals.clear();
  a_timer = NULL;
  a_request = NULL;
}

void ldap::Response::activateTimer()
throw(RuntimeException) {
  a_timer = TimerManager::instantiate().createTimer(this);
}

void ldap::Response::cancelTimer()
throw() {
  if(a_timer != NULL) {
    try {
      TimerManager::instantiate().cancel(a_timer);
    } catch(RuntimeException& ex) {
      ex.trace();
    }

    a_timer = NULL;
  }
}

ldap::Attribute* ldap::Response::createAttribute(const string& name)
throw(RuntimeException) {
  Attribute* result = a_attributes.create();
  result->setName(name);
  return result;
}

string ldap::Response::asString() const
throw() {
  string result("ldap::Response { ");
  result += ClassCode::asString(a_classCode);
  result += functions::asText(" | IdMessage: ", a_idMessage);
  result += " | ";
  result += a_resultCode.asString();
  return result += " }";
}

