// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/time/functions.hpp>
#include <anna/time/internal/Timezone.hpp>

// Standard
#include <stdlib.h> // getenv
#include <time.h>
#include <sys/time.h>
#include <string>

#include <anna/core/functions.hpp>


namespace anna {
namespace time {
Timezone SystemTimezone;
unsigned long long int SecondsReference;
}
}


void anna::time::functions::initialize() throw() {
  static bool cached = false;
  if(!cached) {
    SystemTimezone.set(getenv("TZ"));
    SecondsReference = 0;
    cached = true;
  }
}

void anna::time::functions::setControlPoint(unsigned long long int secondsTimestamp) throw() {
  SecondsReference = secondsTimestamp ? secondsTimestamp : (::time(NULL));
}

unsigned long long int anna::time::functions::getSecondsReference() throw() {
  return SecondsReference;
}

const anna::time::Timezone & anna::time::functions::getSystemTimezone(void) throw() {
  return (SystemTimezone);
}

bool anna::time::functions::initialized(void) throw() {
  return (SystemTimezone.isInitialized());
}

unsigned long long int anna::time::functions::unixSeconds(void) throw() {
  return (::time(NULL));
}

unsigned long long int anna::time::functions::unixMilliseconds() throw() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
}

unsigned long long int anna::time::functions::lapsedMilliseconds() throw() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return ((tv.tv_sec - SecondsReference) * 1000) + (tv.tv_usec / 1000);
}

unsigned long long int anna::time::functions::unixMicroseconds(void) throw() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long long int result(tv.tv_sec);
  result *= 1000000;
  return result += tv.tv_usec;
}

std::string anna::time::functions::currentTimeAsString(void) throw() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  struct tm *tm;
  tm = localtime(&tv.tv_sec);
  return (anna::functions::asString("%02d:%02d:%02d %d usecs", tm->tm_hour, tm->tm_min, tm->tm_sec, tv.tv_usec));
}

