// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>

#include <anna/io/functions.hpp>

#include <libxml/HTMLparser.h>

#include <anna/xml/DocumentFile.hpp>

using namespace std;
using namespace anna;
using namespace anna::xml;

_xmlDoc* DocumentFile::do_initialize(const char* filename)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tf("anna::xml::DocumentFile", "do_initialize", ANNA_FILE_LOCATION));
  _xmlDoc* result;
  a_filename = filename;
  LOGDEBUG(
    string msg("xml::DocumentFile::do_initialize | Filename: ");
    msg += filename;
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(io::functions::exists(filename) == false) {
    string msg("xml::DocumentFile::do_initialize | File: ");
    msg += filename;
    msg += " | Not found";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  result = xmlParseFile(filename);

  if(result == NULL)
    throw RuntimeException(functions::asString("Error analyzing XML document: %s", filename), ANNA_FILE_LOCATION);

  return result;
}

_xmlDoc* DocumentFile::do_initialize(const anna::DataBlock&)
throw(RuntimeException) {
  throw RuntimeException("xml::DocumentFile::do_initialize | Not implemented", ANNA_FILE_LOCATION);
  return NULL;
}


const anna::DataBlock& DocumentFile::getContent() const
throw(RuntimeException) {
  const anna::DataBlock& result = Document::getContent();

  if(result.isEmpty() == false)
    return result;

  int stream;

  if((stream = open(a_filename.c_str(), O_RDONLY)) == -1)
    throw RuntimeException(a_filename, errno, ANNA_FILE_LOCATION);

  const int size = lseek(stream, 0, SEEK_END);
  lseek(stream, 0, SEEK_SET);
  char* buffer = new char [size];
  read(stream, buffer, size);
  close(stream);
  const_cast <DocumentFile*>(this)->setContent(buffer, size);
  delete buffer;
  return result;
}


